# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.


"""
Show a music sample preview.
"""


import os
import re
import tempfile

from PyQt6.QtCore import (
    QSettings,
    Qt
)
from PyQt6.QtWidgets import (
    QComboBox,
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QVBoxLayout,
    QWidget,
)

import app
import appinfo
import musicpreview
import util
import widgets.urlrequester


def get_persistent_cache_dir():
    """
    Determine location for "persistent" caching of music fonts,
    either from the Preference (persistent) or the default temporary
    directory, which will be purged upon computer shutdown.
    """
    pref = QSettings().value('music-fonts/font-cache', '', str)
    return pref or os.path.join(
        tempfile.gettempdir(),
        appinfo.name + '-music-font-samples'
    )


class FontsPreviewWidget(QWidget):
    """Show a preview score using the font selection."""

    # Permanently cache compilations of the provided samples
    persistent_cache_dir = get_persistent_cache_dir()
    # Cache compilations of custom samples for Frescobaldi's lifetime only
    temp_dir = util.tempdir()

    def __init__(self, parent):
        super().__init__(parent)

        # Create the cache directory for default samples
        os.makedirs(self.persistent_cache_dir, 0o700, exist_ok=True)

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        # Label with text "Example:"
        self.lb_sample = QLabel()

        # ComboBox for provided default samples and other options
        self.cb_samples = QComboBox()

        # Select custom file
        self.custom_sample_url = widgets.urlrequester.UrlRequester(
            fileMode=QFileDialog.FileMode.ExistingFile,
            mustExist=True
        )

        # put the default sample in the combobox
        self.populate_default_samples()
        # add other actions
        self.cb_samples.insertSeparator(self.cb_samples.count())
        self.cb_samples.addItem(_("Custom"), "<CUSTOM>")
        self.cb_samples.addItem(_("Current Document"), "<CURRENT>")

        # Add sample source widgets to layout
        bl = QHBoxLayout()
        bl.setContentsMargins(0, 0, 0, 0)
        bl.addWidget(self.lb_sample)
        bl.addWidget(self.cb_samples)
        bl.addWidget(self.custom_sample_url)
        layout.addLayout(bl)

        # The score preview widget
        self.musicFontPreview = mfp = musicpreview.MusicPreviewWidget(
            parent,
            showProgress=False,
            showWaiting=True,
            showLog=False
        )
        layout.addWidget(mfp)

        app.translateUI(self)
        self.loadSettings()

        # Trigger showing of new samples
        self.cb_samples.currentIndexChanged.connect(self.show_sample)
        self.custom_sample_url.editingFinished.connect(self.show_sample)
        parent.finished.connect(self.saveSettings)

    def translateUI(self):
        self.lb_sample.setText("Example:")
        csu = self.custom_sample_url
        csu.setToolTip(_(
            "Use custom sample for music font.\n"
            + "NOTE: This should not include a version statement "
            + "or a \\paper {...} block."
        ))
        csu.setDialogTitle(_("Select sample score"))
        csu.fileDialog(True).setNameFilters(['LilyPond files (*.ly)'])
        i = self.cb_samples.findData("<CURRENT>")
        self.cb_samples.setItemData(i, _(
            "Use current document as music font sample.\n"
            + "NOTE: This is not robust if the document contains "
            + "a \\paper {...} block."
        ), Qt.WindowType.ToolTip)
        i = self.cb_samples.findData("<CUSTOM>")
        self.cb_samples.setItemData(i, csu.toolTip(), Qt.WindowType.ToolTip)

    def loadSettings(self):
        s = QSettings()
        s.beginGroup('document-fonts-dialog')
        sample = s.value('default-music-sample', '', str)
        index = max(0, self.cb_samples.findData(sample))
        self.cb_samples.setCurrentIndex(index)
        custom_sample = s.value('custom-music-sample-url', '', str)
        self.custom_sample_url.setPath(custom_sample)
        if custom_sample:
            sample_dir = os.path.dirname(custom_sample)
        else:
            sample_dir = os.path.dirname(
                self.window().parent().currentDocument().url().toLocalFile())
        self.custom_sample_url.fileDialog().setDirectory(sample_dir)

    def saveSettings(self):
        s = QSettings()
        s.beginGroup('document-fonts-dialog')
        s.setValue('default-music-sample', self.cb_samples.currentData())
        s.setValue('custom-music-sample-url', self.custom_sample_url.path())

    def populate_default_samples(self):
        """Populate hte default samples ComboBox.
        This is just factored out to unclutter __init__.
        """
        cb = self.cb_samples

        def add_entry(entry):
            cb.addItem(entry['label'], entry['file'])
            cb.setItemData(cb.count() - 1, entry['tooltip'], Qt.WindowType.ToolTip)

        add_entry({
            'label': _('Bach (Piano)'),
            'file': 'bach.ly',
            'tooltip': _(
                "Baroque music lends itself to traditional fonts"
            )
        })
        add_entry({
            'label': _('Scriabin (Piano)'),
            'file': 'scriabine.ly',
            'tooltip': _(
                "Late romantic, complex piano music"
            )
        })
        add_entry({
            'label': _('Berg (String Quartet)'),
            'file': 'berg-string-quartet.ly',
            'tooltip': _(
                "Complex score, requires a 'clean' font"
            )
        })
        add_entry({
            'label': _('Real Book (Lead Sheet)'),
            'file': 'realbook.ly',
            'tooltip': _(
                "Jazz-like lead sheet\n"
                + "NOTE: beautiful results rely on appropriate text fonts.\n"
                + "Good choices are \"lilyjazz-text\" for roman and\n"
                + "\"lilyjazz-chords\" for sans text fonts."
            )
        })
        add_entry({
            'label': _('Schenker Diagram'),
            'file': 'schenker.ly',
            'tooltip': _(
                "Schenker diagram with absolutely\n"
                + "non-standard notation"
            )
        })
        add_entry({
            'label': _('Glyphs'),
            'file': 'glyphs.ly',
            'tooltip': _(
                "Non-comprehensive specimen sheet"
            )
        })

    def show_sample(self):
        """Display a sample document for the selected notation font."""
        print("Enter show_sample")
        global_size = ''
        base_dir = None
        sample_content = ''
        cache_persistently = False

        target = self.cb_samples.currentData()
        self.custom_sample_url.setEnabled(target == "<CUSTOM>")

        def handle_staff_size():
            """
            If the sample file *starts with* a staff-size definition
            it will be injected *after* our paper block.
            """
            nonlocal sample_content, global_size
            match = re.match(r'#\(set-global-staff-size \d+\)', sample_content)
            if match:
                global_size = match.group(0)
                sample_content = sample_content[len(global_size):]

        def load_content():
            """
            Load the content to be engraved as sample,
            either from the active editor or from a file.
            """
            nonlocal sample_content, base_dir
            nonlocal cache_persistently, target
            custom_file = self.custom_sample_url.path()

            if target == "<CUSTOM>" and not custom_file:
                target = self.cb_samples.itemData(0)

            # Provided sample files will be cached persistently
            cache_persistently = target not in ("<CUSTOM>", "<CURRENT>")

            if target == "<CURRENT>":
                # Engrave active document
                import engrave
                current_doc = engrave.engraver(app.activeWindow()).document()
                sample_content = current_doc.toPlainText()
                if not current_doc.url().isEmpty():
                    base_dir = os.path.dirname(current_doc.url().toLocalFile())
            else:
                if target == "<CUSTOM>":
                    print("Custom file:", custom_file)
                    sample_file = custom_file
                else:
                    # Engrave from a file
                    import fonts
                    template_dir = os.path.join(fonts.__path__[0], 'templates')
                    sample_file = os.path.join(
                        template_dir,
                        'musicfont-' + target
                    )
                    print("Default:", sample_file)
                base_dir = os.path.dirname(sample_file)
                with open(sample_file) as f:
                    sample_content = f.read()

        def sample_document():
            """
            Steps of composing the used sample document.
            """
            load_content()
            handle_staff_size()
            result = [
                '\\version "{}"\n'.format(
                    self.window().available_fonts.music_fonts(
                    ).lilypond_info.versionString()
                ),
                f'{global_size}\n' if global_size else '',
                # TODO: "Protect" this regarding openLilyLib.
                # It would be easy to simply pass 'lily' as an argument
                # to always use the generic approach. However, that would
                # prevent the use of font extensions and stylesheets.
                self.window().font_full_cmd(),
                sample_content
            ]
            return '\n'.join(result)

        sample = sample_document()
        cache_dir = (
            self.persistent_cache_dir
            if cache_persistently
            else self.temp_dir
        )
        self.musicFontPreview.preview(
            sample,
            title='Music font preview',
            base_dir=base_dir,
            temp_dir=cache_dir,
            cached=True)



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
The browser widget for the help browser.
"""


import os

from PyQt6.QtCore import QSettings, Qt, QUrl
from PyQt6.QtGui import QKeySequence
from PyQt6.QtPrintSupport import QPrintDialog, QPrinter
from PyQt6.QtWebEngineCore import QWebEnginePage, QWebEngineSettings
from PyQt6.QtWebEngineWidgets import QWebEngineView
from PyQt6.QtWidgets import QComboBox, QLineEdit, QMenu, QToolBar, QVBoxLayout, QWidget

import app
import icons
import helpers
import lilypondinfo
import lilydoc.manager
import lilydoc.network
import textformats


class Browser(QWidget):
    """LilyPond documentation browser widget."""
    def __init__(self, dockwidget):
        super().__init__(dockwidget)

        layout = QVBoxLayout(spacing=0)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self.toolbar = tb = QToolBar()
        self.webview = QWebEngineView(self, contextMenuPolicy=Qt.ContextMenuPolicy.CustomContextMenu)
        self.webview.setPage(WebEnginePage(self.webview))
        self.chooser = QComboBox(sizeAdjustPolicy=QComboBox.SizeAdjustPolicy.AdjustToContents)
        self.search = SearchEntry(maximumWidth=200, clearButtonEnabled=True)

        layout.addWidget(self.toolbar)
        layout.addWidget(self.webview)

        ac = dockwidget.actionCollection
        ac.help_back.triggered.connect(self.webview.back)
        ac.help_forward.triggered.connect(self.webview.forward)
        ac.help_home.triggered.connect(self.showHomePage)
        ac.help_web_browser_homepage.triggered.connect(self.openWebBrowserHomePage)
        ac.help_web_browser.triggered.connect(self.openWebBrowser)
        ac.help_print.triggered.connect(self.slotPrint)

        self.webview.urlChanged.connect(self.slotUrlChanged)
        self.webview.customContextMenuRequested.connect(self.slotShowContextMenu)

        tb.addAction(ac.help_back)
        tb.addAction(ac.help_forward)
        tb.addSeparator()
        tb.addAction(ac.help_home)
        tb.addAction(ac.help_web_browser)
        w = tb.widgetForAction(ac.help_web_browser)
        w.addAction(ac.help_web_browser_homepage)
        tb.addAction(ac.help_print)
        tb.addSeparator()
        tb.addWidget(self.chooser)
        tb.addWidget(self.search)

        self.chooser.activated[int].connect(self.showHomePage)
        self.search.textChanged.connect(self.slotSearchChanged)
        self.search.returnPressed.connect(self.slotSearchReturnPressed)
        dockwidget.mainwindow().iconSizeChanged.connect(self.updateToolBarSettings)
        dockwidget.mainwindow().toolButtonStyleChanged.connect(self.updateToolBarSettings)

        app.settingsChanged.connect(self.readSettings)
        self.readSettings()
        self.loadDocumentation()
        self.showInitialPage()
        app.settingsChanged.connect(self.loadDocumentation)
        app.translateUI(self)

    def readSettings(self):
        s = QSettings()
        s.beginGroup("documentation")
        ws = self.webview.page().settings()
        family = s.value("fontfamily", self.font().family(), str)
        size = s.value("fontsize", 16, int)
        ws.setFontFamily(QWebEngineSettings.FontFamily.StandardFont, family)
        ws.setFontSize(QWebEngineSettings.FontSize.DefaultFontSize, size)
        fixed = textformats.formatData('editor').font
        ws.setFontFamily(QWebEngineSettings.FontFamily.FixedFont, fixed.family())
        ws.setFontSize(QWebEngineSettings.FontSize.DefaultFixedFontSize, int(fixed.pointSizeF() * 96 / 72))
        self.webview.page().profile().setHttpAcceptLanguage(','.join(lilydoc.network.langs()))

    def keyPressEvent(self, ev):
        if ev.text() == "/":
            self.search.setFocus()
        else:
            super().keyPressEvent(ev)

    def translateUI(self):
        try:
            self.search.setPlaceholderText(_("Search..."))
        except AttributeError:
            pass # not in Qt 4.6

    def showInitialPage(self):
        """Shows the preferred start page.

        If a local documentation instance already has a suitable version,
        just loads it. Otherwise connects to the allLoaded signal, that is
        emitted when all the documentation instances have loaded their version
        information and then shows the start page (if another page wasn't yet
        loaded).

        """
        if self.webview.url().isEmpty():
            docs = lilydoc.manager.docs()
            version = lilypondinfo.preferred().version()
            index = -1
            if version:
                for num, doc in enumerate(docs):
                    if doc.version() is not None and doc.version() >= version:
                        index = num # a suitable documentation is found
                        break
            if index == -1:
                # nothing found (or LilyPond version not available),
                # wait for loading or show the most recent version
                if not lilydoc.manager.loaded():
                    lilydoc.manager.allLoaded.connect(self.showInitialPage)
                    return
                index = len(docs) - 1
            self.chooser.setCurrentIndex(index)
            self.showHomePage()

    def loadDocumentation(self):
        """Puts the available documentation instances in the combobox."""
        i = self.chooser.currentIndex()
        self.chooser.clear()
        for doc in lilydoc.manager.docs():
            v = doc.versionString()
            if doc.isLocal():
                t = _("(local)")
            else:
                t = _("({hostname})").format(hostname=doc.url().host())
            self.chooser.addItem("{} {}".format(v or _("<unknown>"), t))
        self.chooser.setCurrentIndex(i)
        if not lilydoc.manager.loaded():
            lilydoc.manager.allLoaded.connect(self.loadDocumentation, -1)
            return

    def updateToolBarSettings(self):
        mainwin = self.parentWidget().mainwindow()
        self.toolbar.setIconSize(mainwin.iconSize())
        self.toolbar.setToolButtonStyle(mainwin.toolButtonStyle())

    def showManual(self):
        """Invoked when the user presses F1."""
        self.slotHomeFrescobaldi() # TEMP

    def slotUrlChanged(self):
        ac = self.parentWidget().actionCollection
        ac.help_back.setEnabled(self.webview.history().canGoBack())
        ac.help_forward.setEnabled(self.webview.history().canGoForward())

    def openUrl(self, url):
        if url.path().endswith(('.ily', '.lyi', '.ly')):
            self.sourceViewer().showReply(lilydoc.network.get(url))
        else:
            self.webview.load(url)

    def slotUnsupported(self, reply):
        helpers.openUrl(reply.url())

    def slotSearchChanged(self):
        text = self.search.text()
        if not text.startswith(':'):
            self.webview.page().findText(text)

    def slotSearchReturnPressed(self):
        text = self.search.text()
        if not text.startswith(':'):
            self.slotSearchChanged()
        else:
            pass # TODO: implement full doc search

    def sourceViewer(self):
        try:
            return self._sourceviewer
        except AttributeError:
            from . import sourceviewer
            self._sourceviewer = sourceviewer.SourceViewer(self)
            return self._sourceviewer

    def getHomePageUrl(self):
        """Returns the URL of the LilyPond documentation."""
        i = self.chooser.currentIndex()
        if i < 0:
            i = 0
        doc = lilydoc.manager.docs()[i]

        url = doc.home()
        if doc.isLocal():
            path = url.toLocalFile()
            langs = lilydoc.network.langs()
            if langs:
                for lang in langs:
                    if os.path.exists(path + '.' + lang + '.html'):
                        path += '.' + lang
                        break
            url = QUrl.fromLocalFile(path + '.html')
        return url

    def showHomePage(self):
        self.webview.load(self.getHomePageUrl())

    def openWebBrowser(self):
        helpers.openUrl(self.webview.url())

    def openWebBrowserHomePage(self):
        helpers.openUrl(self.getHomePageUrl())

    def slotPrint(self):
        printer = self._printer = QPrinter()
        dlg = QPrintDialog(printer, self)
        dlg.setWindowTitle(app.caption(_("Print")))
        if dlg.exec():
            self.webview.page().print(printer, self.slotPrintingDone)

    def slotPrintingDone(self, success):
        del self._printer

    def slotShowContextMenu(self, pos):
        d = self.webview.lastContextMenuRequest()
        menu = QMenu()
        if d.linkUrl().isValid():
            a = self.webview.pageAction(QWebEnginePage.WebAction.CopyLinkToClipboard)
            a.setIcon(icons.get("edit-copy"))
            a.setText(_("Copy &Link"))
            menu.addAction(a)
            menu.addSeparator()
            a = menu.addAction(icons.get("internet-web-browser"), _("Open Link in Web Browser"))
            a.triggered.connect((lambda url: lambda: self.slotNewWindow(url))(d.linkUrl()))
        else:
            if d.selectedText():
                a = self.webview.pageAction(QWebEnginePage.WebAction.Copy)
                a.setIcon(icons.get("edit-copy"))
                a.setText(_("&Copy"))
                menu.addAction(a)
                menu.addSeparator()
            a = menu.addAction(icons.get("internet-web-browser"), _("Open Current Page in Web Browser"))
            a.triggered.connect((lambda url: lambda: self.slotNewWindow(url))(self.webview.url()))
        if menu.actions():
            menu.exec(self.webview.mapToGlobal(pos))

    def slotNewWindow(self, url):
        helpers.openUrl(url)


class SearchEntry(QLineEdit):
    """A line edit that clears itself when ESC is pressed."""
    def keyPressEvent(self, ev):
        if ev.key() == Qt.Key.Key_Escape:
            if self.text():
                self.clear()
            else:
                webview = self.parentWidget().parentWidget().webview
                webview.setFocus()
                webview.page().findText(None)
        elif any(ev.matches(key) for key in (
            QKeySequence.StandardKey.MoveToNextLine, QKeySequence.StandardKey.MoveToPreviousLine,
            QKeySequence.StandardKey.MoveToNextPage, QKeySequence.StandardKey.MoveToPreviousPage,
                )):
            webview = self.parentWidget().parentWidget().webview
            webview.keyPressEvent(ev)
        else:
            super().keyPressEvent(ev)


class WebEnginePage(QWebEnginePage):
    """QWebEnginePage that shows LY files using the source viewer."""
    def acceptNavigationRequest(self, url, type, isMainFrame):
        if url.path().endswith(('.ily', '.lyi', '.ly')):
            self.parent().parent().sourceViewer().showReply(lilydoc.network.get(url))
            return False
        return super().acceptNavigationRequest(url, type, isMainFrame)



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
A QWidget showing a "could not load PyQtWebEngine" message.
"""


import platform

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QLabel, QLayout, QVBoxLayout, QWidget

import app


class WebEngineDummy(QWidget):
    def __init__(self, dockwidget):
        super().__init__(dockwidget)
        self._msg = QLabel(openExternalLinks = True)
        self._msg.setTextFormat(Qt.TextFormat.RichText)
        self._msg.setWordWrap(True)
        # take into account docbrowser.HelpBrowser.activate()
        self.webview = self._msg
        layout = QVBoxLayout()
        self.setLayout(layout)
        layout.addWidget(self._msg)
        app.translateUI(self)

    def translateUI(self):
        message = (_("<b>Could not load PyQtWebEngine, "
                     + "so the {tool} cannot be loaded.</b>").format(
                     tool = self.parentWidget().windowTitle()))
        if platform.system() == "Darwin":
            import macos
            if (macos.inside_app_bundle()
                    and not macos.inside_lightweight_app_bundle()):
                message = (
                    message + "<br><br>"
                    + _("PyQtWebEngine has been intentionally excluded from "
                        + "the application bundle: at the moment, including "
                        + "it results in a faulty application.<br><br>"
                        + "For more details, see <a href=\"{issue}\">issue #1244</a> "
                        + "on Frescobaldi's GitHub page.<br>"
                        + "For more information on alternative ways of running "
                        + "Frescobaldi, see the <a href=\"{wiki}\">relevant Wiki "
                        + "page</a> on GitHub.").format(
                        issue = "https://github.com/frescobaldi/frescobaldi/issues/1244",
                        wiki = "https://github.com/frescobaldi/frescobaldi/"
                               + "wiki/How-to-install-Frescobaldi-on-Mac-OS-X"))
        self._msg.setText(message)



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Extensions preferences.
"""

import re

from PyQt6.QtCore import QSettings, Qt
from PyQt6.QtGui import QFont, QStandardItem, QStandardItemModel

from PyQt6.QtWidgets import (
    QAbstractItemView,
    QCheckBox,
    QHBoxLayout,
    QHeaderView,
    QLabel,
    QTreeView,
    QVBoxLayout,
    QWidget
)

import app
import appinfo
import preferences
from widgets import urlrequester
from extensions import FailedTree

class Extensions(preferences.ScrolledGroupsPage):
    """Extensions page for Preferences"""

    def __init__(self, dialog):
        super().__init__(dialog)

        self._active = QSettings().value(
            'extension-settingss/active', True, bool)

        layout = QVBoxLayout()
        self.scrolledWidget.setLayout(layout)

        layout.addWidget(General(self))
        layout.addWidget(Installed(self))
        if FailedTree._model:
            layout.addWidget(Failed(self))
        layout.addWidget(Config(self))
        layout.addStretch(1)

    def active(self):
        return self._active


class General(preferences.Group):
    """Application-wide settings for extensions."""

    def __init__(self, page):
        super().__init__(page)

        layout = QVBoxLayout()
        self.setLayout(layout)

        header_layout = QHBoxLayout()
        self.active = QCheckBox(toggled=self.changed)
        header_layout.addWidget(self.active)

        self.api_version = QLabel()
        header_layout.addStretch()
        header_layout.addWidget(self.api_version)

        layout.addLayout(header_layout)

        self.root = urlrequester.UrlRequester()
        self.root.changed.connect(self.changed)
        layout.addWidget(self.root)

        app.translateUI(self)

    def translateUI(self):
        self.setTitle(_("General Settings"))
        self.active.setText(_("Use Extensions"))
        self.api_version.setText(_("Extension API: {apiversion}").format(
            apiversion=appinfo.extension_api))
        self.active.setToolTip(_("If unchecked don't look for extensions."))

    def loadSettings(self):
        s = QSettings()
        s.beginGroup("extension-settings")
        self.active.setChecked(s.value("active", True, bool))
        self.root.setPath(s.value("root-directory", '', str))

    def saveSettings(self):
        s = QSettings()
        s.beginGroup("extension-settings")
        s.setValue("active", self.active.isChecked())
        s.setValue("root-directory", self.root.path())

    def root_changed(self):
        self.changed.emit()


class Installed(preferences.Group):
    """Overview of installed extensions.

    A QTreeView lists the metadata of all installed extensions.
    If the currently selected extension provides a configuration
    widget it is displayed in the bottom group of the page.

    With a checkbox individual extensions can be deactivated.
    Metadata is listed for all *installed* extensions, regardless
    of manual deactivation or load failure.
    """

    def __init__(self, page):
        super().__init__(page)

        layout = QVBoxLayout()
        self.setLayout(layout)

        # This must be called before self.populate() because
        # the data model relies on the labels
        app.translateUI(self)

        self.tree = QTreeView()
        self.name_items = {}
        self._selected_extension = ''
        self.tree.setModel(QStandardItemModel())
        self.tree.model().setColumnCount(2)
        self.tree.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.tree.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.tree.setHeaderHidden(True)
        self.tree.header().setSectionResizeMode(
            0, QHeaderView.ResizeMode.ResizeToContents)
        self.tree.selectionModel().selectionChanged.connect(
            self.selection_changed)
        self.populate()
        layout.addWidget(self.tree)

    def translateUI(self):
        self.setTitle(_("Installed Extensions"))
        self.config_labels = {
            'extension-name': _("Name"),
            'maintainers': _("Maintainer(s)"),
            'version': _("Version"),
            'api-version': _("API version"),
            'license': _("License"),
            'short-description': _("Short Description"),
            'description': _("Description"),
            'repository': _("Repository"),
            'website': _("Website"),
            'dependencies': _("Dependencies")
        }

    def loadSettings(self):
        s = QSettings()
        self.setEnabled(self.page().active())
        s.beginGroup("extension-settings/installed")
        inactive = s.value("inactive", [], list)
        for ext in self.name_items.keys():
            self.name_items[ext].setCheckState(
                Qt.CheckState.Checked if ext not in inactive else Qt.CheckState.Unchecked)
        self.tree.model().dataChanged.connect(self.page().changed)

    def saveSettings(self):
        s = QSettings()
        s.beginGroup("extension-settings/installed")
        inactive = [ext for ext in self.name_items.keys()
            if self.name_items[ext].checkState() == Qt.CheckState.Unchecked]
        s.setValue("inactive", inactive)

    def populate(self):
        """Populate the tree view with data from the installed extensions.
        """

        # TODO/Question:
        # Would it make sense to move this to a dedicated model class
        # complementing the FailedModel?

        root = self.tree.model().invisibleRootItem()
        extensions = app.extensions()
        for ext in extensions.installed_extensions():
            ext_infos = extensions.infos(ext)
            display_name = ext_infos.get(ext, ext) if ext_infos else ext.name()
            loaded_extension = extensions.get(ext)
            if loaded_extension:
                display_name += f' ({loaded_extension.load_time()})'

            name_item = QStandardItem(display_name)
            name_item.extension_name = ext
            name_item.setCheckable(True)
            self.name_items[ext] = name_item
            icon = extensions.icon(ext)
            if icon:
                name_item.setIcon(icon)
            root.appendRow([name_item])
            for entry in [
                'extension-name',
                'short-description',
                'description',
                'version',
                'api-version',
                'dependencies',
                'maintainers',
                'repository',
                'website',
                'license'
            ]:
                label_item = QStandardItem('{}:'.format(
                    self.config_labels[entry]))
                label_item.setTextAlignment(Qt.AlignmentFlag.AlignTop)
                bold = QFont()
                bold.setWeight(QFont.Weight.Bold)
                label_item.setFont(bold)
                details = ext_infos.get(entry, "") if ext_infos else ""
                if type(details) is list:
                    details = '\n'.join(details)
                details_item = QStandardItem(details)
                details_item.setTextAlignment(Qt.AlignmentFlag.AlignTop)
                if entry == 'api-version':
                    # Check for correct(ly formatted) api-version entry
                    # and highlight it in case of mismatch
                    api_version = appinfo.extension_api
                    if not details:
                        details_item.setFont(bold)
                        details_item.setText(
                            _("Misformat: {api}").format(details))
                    elif not details == api_version:
                            details_item.setFont(bold)
                            details_item.setText('{} ({}: {})'.format(
                                details,
                                appinfo.appname,
                                api_version))
                name_item.appendRow([label_item, details_item])

    def selected_extension(self):
        """Return the (directory) name of the extension that
        is currently selected."""
        return self._selected_extension

    def selection_changed(self, new, old):
        """Show the configuration widget for the selected extension,
        if available."""
        config = self.siblingGroup(Config)
        if new.indexes():
            ext_item = self.tree.model().itemFromIndex(new.indexes()[0])
            # NOTE: This may have to be changed if there should be
            # more complexity in the tree model than now (a selected
            # row is either a top-level row or its immediate child)
            if not hasattr(ext_item, 'extension_name'):
                ext_item = ext_item.parent()
            name = ext_item.extension_name
            if name == self.selected_extension():
                return
        else:
            # If nothing is selected, show the "empty" widget
            name = ""

        config.hide_extension()
        self._selected_extension = name
        config.show_extension(name)


class Failed(preferences.Group):
    """Show information about extensions that failed to load properly.
    This is only instantiated if there *are* failed extensions."""

    def __init__(self, page):
        super().__init__(page)
        import extensions
        layout = QVBoxLayout()
        self.setLayout(layout)
        self.tree = extensions.FailedTree(self)
        layout.addWidget(self.tree)
        app.translateUI(self)

    def translateUI(self):
        self.setTitle(_("Failed Extensions"))


class Config(preferences.Group):
    """Configuration of an extension.
    If the currently selected extension doesn't provide a
    config widget this will show a dummy widget instead."""

    def __init__(self, page):
        super().__init__(page)

        layout = QVBoxLayout()
        self.setLayout(layout)

        self._widgets = {}
        # Create an "empty" widget as a placeholder for
        # extensions that don't provide a config widget
        self.empty_widget = QWidget(self)
        empty_layout = QVBoxLayout()
        self.empty_widget.setLayout(empty_layout)
        self.empty_label = QLabel(self)
        empty_layout.addWidget(self.empty_label)
        layout.addWidget(self.empty_widget)

        # Reference to the current widget, initialize to empty widget
        self.current_widget = self.empty_widget
        self.current_widget.hide()

        app.translateUI(self)

    def translateUI(self):
        self.setTitle(_("Extension Configuration"))
        self.empty_label.setText(_(
            "The selected extension does not "
            "provide a configuration widget "
            "or is inactive."))

    def loadSettings(self):
        """Load global settings. Any ConfigWidget class must implement
        save_settings(), but this is called right after instantiating
        the widget in self.widget()."""
        self.setEnabled(self.page().active())

    def saveSettings(self):
        """Ask all extension configuration widgets to save their settings.
        Configuration widgets are required to implement save_settings()."""
        for widget in self._widgets.values():
            if widget:
                widget.save_settings()

    def hide_extension(self):
        """Remove the currently displayed configuration widget."""
        self.layout().removeWidget(self.current_widget)
        self.current_widget.hide()
        self.current_widget = None

    def show_extension(self, name):
        """Show an extension's configuration widget, or the
        empty widget if the extension does not provide one."""
        #widget = self._widgets.get(name, self.empty_widget)
        widget = self.widget(name) or self.empty_widget
        self.current_widget = widget
        self.layout().addWidget(widget)
        widget.show()

    def widget(self, extension):
        """Return a configuration widget if provided by the extension,
        or None otherwise."""
        widget = self._widgets.get(extension, False)
        if not widget:
            ext = app.extensions().get(extension)
            # skip non-loaded extensions
            if ext:
                widget = ext.config_widget(self)
                if widget:
                    widget.load_settings()
            self._widgets[extension] = widget or None
        return widget



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Custom engraving dialog.
"""


import os
import collections

from PyQt6.QtCore import QSettings, QSize, Qt
from PyQt6.QtWidgets import (QCheckBox, QComboBox, QDialog, QDialogButtonBox,
    QGridLayout, QLabel, QSpinBox, QTextEdit)

import ly.xml

import app
import userguide
import icons
import job
import panelmanager
import lilychooser
import listmodel
import widgets
import qutil
import util


class Dialog(QDialog):
    def __init__(self, mainwindow):
        super().__init__(mainwindow)
        self._document = None

        layout = QGridLayout()
        self.setLayout(layout)

        self.versionLabel = QLabel()
        self.lilyChooser = lilychooser.LilyChooser()

        self.outputLabel = QLabel()
        self.outputCombo = QComboBox()

        self.resolutionLabel = QLabel()
        self.resolutionCombo = QComboBox(editable=True)

        self.antialiasLabel = QLabel()
        self.antialiasSpin = QSpinBox(minimum=1, maximum=128, value=1)

        self.modeLabel = QLabel()
        self.modeCombo = QComboBox()

        self.deleteCheck = QCheckBox()
        self.embedSourceCodeCheck = QCheckBox()
        self.englishCheck = QCheckBox()

        self.commandLineLabel = QLabel()
        self.commandLine = QTextEdit(acceptRichText=False)

        self.buttons = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        self.buttons.button(QDialogButtonBox.StandardButton.Ok).setIcon(icons.get("lilypond-run"))
        userguide.addButton(self.buttons, "engrave_custom")

        self.resolutionCombo.addItems(['100', '200', '300', '600', '1200'])
        self.resolutionCombo.setCurrentIndex(2)

        self.modeCombo.addItems(['preview', 'publish', 'debug', 'incipit'])
        layout.addWidget(self.versionLabel, 0, 0)
        layout.addWidget(self.lilyChooser, 0, 1, 1, 3)
        layout.addWidget(self.outputLabel, 1, 0)
        layout.addWidget(self.outputCombo, 1, 1, 1, 3)
        layout.addWidget(self.resolutionLabel, 2, 0)
        layout.addWidget(self.resolutionCombo, 2, 1)
        layout.addWidget(self.antialiasLabel, 2, 2, Qt.AlignmentFlag.AlignRight)
        layout.addWidget(self.antialiasSpin, 2, 3)
        layout.addWidget(self.modeLabel, 3, 0)
        layout.addWidget(self.modeCombo, 3, 1, 1, 3)
        layout.addWidget(self.deleteCheck, 4, 0, 1, 4)
        layout.addWidget(self.embedSourceCodeCheck, 5, 0, 1, 4)
        layout.addWidget(self.englishCheck, 6, 0, 1, 4)
        layout.addWidget(self.commandLineLabel, 7, 0, 1, 4)
        layout.addWidget(self.commandLine, 8, 0, 1, 4)
        layout.addWidget(widgets.Separator(), 9, 0, 1, 4)
        layout.addWidget(self.buttons, 10, 0, 1, 4)

        app.translateUI(self)
        qutil.saveDialogSize(self, "engrave/custom/dialog/size", QSize(480, 260))
        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)

        model = listmodel.ListModel(formats, display=lambda f: f.title(),
            icon=lambda f: icons.file_type(f.type))
        self.outputCombo.setModel(model)

        s = QSettings()
        s.beginGroup("lilypond_settings")
        self.englishCheck.setChecked(
            s.value("no_translation", False, bool))
        self.deleteCheck.setChecked(
            s.value("delete_intermediate_files", True, bool))

        if s.value("default_output_target", "pdf", str) == "svg":
            self.outputCombo.setCurrentIndex(3)

        app.jobFinished.connect(self.slotJobFinished)
        self.outputCombo.currentIndexChanged.connect(self.updateFormatControls)
        self.updateFormatControls()

    def translateUI(self):
        self.setWindowTitle(app.caption(_("Engrave custom")))
        self.versionLabel.setText(_("LilyPond Version:"))
        self.outputLabel.setText(_("Output Format:"))
        self.resolutionLabel.setText(_("Resolution:"))
        self.antialiasLabel.setText(_("Antialias Factor:"))
        self.modeLabel.setText(_("Engraving mode:"))
        self.modeCombo.setItemText(0, _("Preview"))
        self.modeCombo.setItemText(1, _("Publish"))
        self.modeCombo.setItemText(2, _("First System Only"))
        self.modeCombo.setItemText(3, _("Layout Control"))
        self.deleteCheck.setText(_("Delete intermediate output files"))
        self.embedSourceCodeCheck.setText(_("Embed Source Code"))
        self.englishCheck.setText(_("Run LilyPond with English messages"))
        self.commandLineLabel.setText(_("Additional Command Line Options:"))
        self.buttons.button(QDialogButtonBox.StandardButton.Ok).setText(_("Run LilyPond"))
        self.outputCombo.update()

    def slotJobFinished(self, doc):
        if doc == self._document:
            self.buttons.button(QDialogButtonBox.StandardButton.Ok).setEnabled(True)
            self._document = None

    def setDocument(self, doc):
        j = job.manager.job(doc)
        if j:
            self.lilyChooser.setLilyPondInfo(j.lilypond_info)
        if j and j.is_running() and not job.attributes.get(j).hidden:
            self._document = doc
            self.buttons.button(QDialogButtonBox.StandardButton.Ok).setEnabled(False)

    def updateFormatControls(self):
        """Reads the widgets and builds a command line."""
        f = formats[self.outputCombo.currentIndex()]
        self.resolutionCombo.setEnabled('resolution' in f.widgets)
        self.antialiasSpin.setEnabled('antialias' in f.widgets)

    def getJob(self, document):
        """Returns and configures a Job to start."""
        f = formats[self.outputCombo.currentIndex()]
        d_options = {}
        args = []

        # Configure job type, choose class
        if self.modeCombo.currentIndex() == 0:   # preview mode
            job_class = job.lilypond.PreviewJob
        elif self.modeCombo.currentIndex() == 1: # publish mode
            job_class = job.lilypond.PublishJob
        elif self.modeCombo.currentIndex() == 2: # incipit mode
            job_class = job.lilypond.LilyPondJob
            d_options['preview'] = True
            d_options['print-pages'] = False
        else:                                    # debug mode
            job_class = job.lilypond.LayoutControlJob
            args = panelmanager.manager(
                self.parent()).layoutcontrol.widget().preview_options()

        # Instantiate Job
        j = job_class(document, args)
        j.lilypond_info = self.lilyChooser.lilyPondInfo()

        # Configure extended command line options
        d_options['delete-intermediate-files'] = True if self.deleteCheck.isChecked() else False

        if self.embedSourceCodeCheck.isChecked():
            d_options['embed-source-code'] = True

        # Assign extended command line options
        for k in d_options:
            j.set_d_option(k, d_options[k])

        # Determine options for the output target/backend
        d = {
            'version': j.lilypond_info.version,
            'resolution': self.resolutionCombo.currentText(),
            'antialias': self.antialiasSpin.value(),
            'xml_export_init': os.path.join(ly.xml.__path__[0], 'xml-export-init.ly'),
        }
        j.set_backend_args(f.options(d))

        # Parse additional/custom tokens from the text edit
        for t in self.commandLine.toPlainText().split():
            if t.startswith('-d'):
                k, v = job.lilypond.parse_d_option(t)
                j.set_d_option(k, v)
            else:
                j.add_argument(t)

        # Set environment variables for the job
        if self.englishCheck.isChecked():
            j.environment['LC_MESSAGES'] = 'C'
        else:
            j.environment.pop('LC_MESSAGES', None)
        j.set_title("{} {} [{}]".format(
            os.path.basename(j.lilypond_info.command),
                j.lilypond_info.versionString(), document.documentName()))
        return j

    def keyPressEvent(self, ev):
        if ev.key() == Qt.Key.Key_Return and ev.modifiers() == Qt.KeyboardModifier.ControlModifier:
            self.accept()
        else:
            super().keyPressEvent(ev)


Format = collections.namedtuple("Format", "type title options widgets")

formats = [
    Format(
        "pdf",
        lambda: _("PDF"),
        lambda d: ['--pdf'],
        (),
    ),
    Format(
        "ps",
        lambda: _("PostScript"),
        lambda d: ['--ps'],
        (),
    ),
    Format(
        "png",
        lambda: _("PNG"),
        lambda d: [
            '--png',
            '-dresolution={resolution}'.format(**d),
            '-danti-alias-factor={antialias}'.format(**d),
        ],
        ('resolution', 'antialias'),
    ),
    Format(
        "svg",
        lambda: _("SVG"),
        lambda d: ['-dbackend=svg'],
        (),
    ),
    Format(
        "pdf",
        lambda: _("PDF (EPS Backend)"),
        lambda d: ['--pdf', '-dbackend=eps'],
        (),
    ),
    Format(
        "eps",
        lambda: _("Encapsulated PostScript (EPS Backend)"),
        lambda d: ['--ps', '-dbackend=eps'],
        (),
    ),
    Format(
        "png",
        lambda: _("PNG (EPS Backend)"),
        lambda d: [
            '--png',
            '-dbackend=eps',
            '-dresolution={resolution}'.format(**d),
            '-danti-alias-factor={antialias}'.format(**d),
        ],
        ('resolution', 'antialias'),
    ),
    Format(
        "xml",
        lambda: _("XML (test ly.xml module)"),
        lambda d: ['--init', d['xml_export_init'], '-dxml-export=-'],
        (),
    ),
]



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Save the current document as a snippet that appears in File->New from Template.
"""


import app
import documentinfo
import widgets.dialog

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import (
    QCheckBox, QCompleter, QVBoxLayout, QMessageBox, QWidget)

from . import model
from . import snippets


class TemplateDialog(widgets.dialog.TextDialog):
    def __init__(self, parent):
        self._lineEdit = None
        super().__init__(parent)
        self.setWindowTitle(app.caption(_("Save as Template")))
        self.setMessage(_("Please enter a template name:"))
        self.setMinimumWidth(320)
        self.setValidateRegExp(r"\w(.*\w)?")
        e = self._lineEdit = self.mainWidget()
        w = QWidget()
        self.setMainWidget(w)
        c = self._runCheck = QCheckBox(
            _("Run LilyPond when creating a new document from this template"))
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        w.setLayout(layout)
        layout.addWidget(e)
        layout.addWidget(c)
        e.setFocus()

    def lineEdit(self):
        """Return the QLineEdit widget."""
        return self._lineEdit or self.mainWidget()

    def runCheck(self):
        """Return the Run LilyPond checkbox."""
        return self._runCheck


def save(mainwindow):

    titles = {snippets.title(name): name
                  for name in model.model().names()
                  if 'template' in snippets.get(name).variables}

    # would it make sense to run LilyPond after creating a document from this
    # template?
    cursor = mainwindow.textCursor()
    template_run = False
    if documentinfo.mode(cursor.document()) == 'lilypond':
        dinfo = documentinfo.docinfo(cursor.document())
        if dinfo.complete() and dinfo.has_output():
            template_run = True

    dlg = TemplateDialog(mainwindow)
    c = QCompleter(sorted(titles), dlg.lineEdit())
    dlg.lineEdit().setCompleter(c)
    dlg.runCheck().setChecked(template_run)

    result = dlg.exec()
    dlg.deleteLater()
    if not result:
        return # cancelled

    title = dlg.text()
    template_run = dlg.runCheck().isChecked()

    if title in titles:
        if QMessageBox.critical(mainwindow,
            _("Overwrite Template?"),
            _("A template named \"{name}\" already exists.\n\n"
              "Do you want to overwrite it?").format(name=title),
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.Cancel) != QMessageBox.StandardButton.Yes:
            return
        name = titles[title]
    else:
        name = None

    # get the text and insert cursor position or selection
    text = cursor.document().toPlainText()

    repls = [(cursor.position(), '${CURSOR}')]
    if cursor.hasSelection():
        repls.append((cursor.anchor(), '${ANCHOR}'))
        repls.sort()

    result = []
    prev = 0
    for pos, what in repls:
        result.append(text[prev:pos].replace('$', '$$'))
        result.append(what)
        prev = pos
    result.append(text[prev:].replace('$', '$$'))
    text = ''.join(result)

    # add header line, if desired enable autorun
    headerline = '-*- template; indent: no;'
    if template_run:
        headerline += ' template-run;'
    text = headerline + '\n' + text

    # save the new snippet
    model.model().saveSnippet(name, text, title)





# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2019 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Customization classes for the qpageview.View.

This is used throughout Frescobaldi, to obey color settings etc.

"""

import itertools
import os
import platform

from PyQt6.QtCore import pyqtSignal, QMargins, QSettings, Qt
from PyQt6.QtWidgets import QMessageBox
from PyQt6.QtPrintSupport import QPrinter, QPrintDialog

import app
import icons
import textformats
import qpageview
import qpageview.view
import qpageview.layout
import qpageview.printing
import qpageview.magnifier
import qpageview.viewactions
import qpageview.widgetoverlay


class Magnifier(qpageview.magnifier.Magnifier):
    """A magnifier that reads and stores its settings."""
    def __init__(self):
        super().__init__()
        app.settingsChanged.connect(self.readSettings)
        self.readSettings()

    def readSettings(self):
        s = QSettings()
        s.beginGroup("musicview/magnifier")
        size = s.value("size", 350, int)
        if self.MIN_SIZE <= size <= self.MAX_SIZE:
            self.resize(size, size)
        scale = s.value("scalef", 3.0, float)
        super().setScale(scale)

    def setScale(self, scale):
        """Reimplemented to save the scale."""
        super().setScale(scale)
        QSettings().setValue("musicview/magnifier/scalef", scale)

    def resizeEvent(self, ev):
        """Reimplemented to save the size."""
        super().resizeEvent(ev)
        QSettings().setValue("musicview/magnifier/size", ev.size().width())


class PagedView(qpageview.widgetoverlay.WidgetOverlayViewMixin, qpageview.View):
    """A View based on qpageview.View.

    This View has additional features and customisation needed in Frescobaldi.

    """

    MAX_ZOOM = 8.0  # see issue #2035; this should be plenty for our needs

    def __init__(self, parent=None):
        super().__init__(parent)
        self._printer = None
        self.documentPropertyStore = qpageview.view.DocumentPropertyStore()
        self.setMagnifier(Magnifier())
        app.settingsChanged.connect(self.readSettings)
        self.readSettings()

    def readSettings(self):
        # strict paging with pageup/pagedown
        self.strictPagingEnabled = QSettings().value("musicview/strict_paging", False, bool)

        # shadow/margin
        shadow = QSettings().value("musicview/shadow", True, bool)
        oldshadow, self.dropShadowEnabled = self.dropShadowEnabled, shadow
        self.pageLayout().setMargins(QMargins(6, 6, 6, 6) if shadow else QMargins(1, 1, 1, 1))
        changed = shadow != oldshadow

        # kinetic scrolling
        kinetic = QSettings().value("musicview/kinetic_scrolling", True, bool)
        self.kineticPagingEnabled = kinetic
        self.kineticScrollingEnabled = kinetic

        # scrollbar visibility
        scrollbars = QSettings().value("musicview/show_scrollbars", True, bool)
        policy = Qt.ScrollBarPolicy.ScrollBarAsNeeded if scrollbars else Qt.ScrollBarPolicy.ScrollBarAlwaysOff
        oldpolicy = self.horizontalScrollBarPolicy()
        self.setHorizontalScrollBarPolicy(policy)
        self.setVerticalScrollBarPolicy(policy)
        if policy != oldpolicy:
            changed = True

        if not self.pageCount():
            return

        # if margin or scrollbar visibility was changed, relayout
        if changed:
            if self._viewMode:
                self.fitPageLayout()
            self.updatePageLayout()


        # set certain preferences to the existing renderers
        renderers = {page.renderer for page in self.pageLayout() if page.renderer}
        pages = {page for page in self.pageLayout() if not page.renderer}
        changed = False

        # paper color; change the existing renderer
        paperColor = textformats.formatData('editor').baseColors['paper']
        # if a page has no renderer, adjust the page itself :-)
        for rp in itertools.chain(renderers, pages):
            if rp.paperColor != paperColor:
                changed = True
                rp.paperColor = paperColor

        if changed:
            self.rerender()

    def loadPdf(self, filename, renderer=None):
        """Reimplemented to use a customized renderer by default."""
        super().loadPdf(filename, renderer or getRenderer("pdf"))

    def loadSvgs(self, filenames, renderer=None):
        """Reimplemented to use a customized renderer by default."""
        super().loadSvgs(filenames, renderer or getRenderer("svg"))

    def loadImages(self, filenames, renderer=None):
        """Reimplemented to use a customized renderer by default."""
        super().loadImages(filenames, renderer or getRenderer("image"))

    def print(self, printer=None, pageNumbers=None, showDialog=True):
        """Print the contents of the View."""
        if printer is None:
            if self._printer is None:
                self._printer = QPrinter()
            printer = self._printer
        printer.setCopyCount(1) # prevent embarrassing situations :-)
        if self.document() and self.document().filename():
            filename = os.path.basename(self.document().filename())
        else:
            filename = ""
        printer.setDocName(filename)
        if showDialog:
            dlg = QPrintDialog(printer, self)
            if filename:
                title = app.caption(_("Print {filename}").format(filename=filename))
            else:
                title = app.caption(_("Print"))
            dlg.setWindowTitle(title)
            dlg.setMinMax(1, self.pageCount())
            if not dlg.exec():
                return  # cancelled
        s = QSettings()
        printer.setResolution(s.value("printing/dpi", 300, int))

        job = super().print(printer, pageNumbers, False)
        if job:
            progress = PrintProgressDialog(job, self)
            progress.setWindowTitle(title)
            progress.setLabelText(_("Preparing to print..."))
            progress.show()


class PrintProgressDialog(qpageview.printing.PrintProgressDialog):
    def showProgress(self, page, num, total):
        """Called by the job when printing a page."""
        self.setValue(num)
        self.setLabelText(_("Printing page {page} ({num} of {total})...").format(
                page=page, num=num, total=total))

    def showErrorMessage(self):
        """Reimplemented to show a translated error message."""
        QMessageBox.warning(self.parent(), _("Printing Error"),
                    _("Could not send the document to the printer."))


class ViewActions(qpageview.viewactions.ViewActions):
    """View actions with translated texts and icons."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        app.translateUI(self)

    def createActions(self):
        """Creates the actions; called by __init__()."""
        super().createActions()
        # limit the maximum zoom factor to what our PagedView allows
        self.zoomer._zoomFactors = tuple(filter(
            lambda x: x <= PagedView.MAX_ZOOM, self.zoomer._zoomFactors))

    def translateUI(self):
        """Translate our actions correctly."""
        self.setActionTexts(_)
        self.pager.setDisplayFormat(_("{num} of {total}"))

    def setActionIcons(self):
        """Set icons to our actions."""
        self.print.setIcon(icons.get('document-print'))
        self.zoom_in.setIcon(icons.get('zoom-in'))
        self.zoom_out.setIcon(icons.get('zoom-out'))
        self.zoom_original.setIcon(icons.get('zoom-original'))
        self.fit_width.setIcon(icons.get('zoom-fit-width'))
        self.fit_height.setIcon(icons.get('zoom-fit-height'))
        self.fit_both.setIcon(icons.get('zoom-fit-best'))
        self.rotate_left.setIcon(icons.get('rotate-left'))
        self.rotate_right.setIcon(icons.get('rotate-right'))
        self.next_page.setIcon(icons.get('go-next'))
        self.previous_page.setIcon(icons.get('go-previous'))
        self.magnifier.setIcon(icons.get('zoom-magnifier'))
        self.reload.setIcon(icons.get('view-refresh'))


def getRenderer(rendertype):
    """Factory to create a renderer with the paper color adjusted to the prefs.

    Currently, the rendertype can be "pdf", "svg", "image" or "diff".

    For the PDF type, the render backends are also read from the prefs.
    For image types, the paper color is not used. For diff types, the paper-
    color is not set.

    """
    if rendertype == "pdf":
        import qpageview.pdf
        r = qpageview.pdf.PdfRenderer()
    elif rendertype == "svg":
        import qpageview.svg
        r = qpageview.svg.SvgRenderer()
    elif rendertype == "image":
        import qpageview.image
        r = qpageview.image.ImageRenderer()
    elif rendertype == "diff":
        import qpageview.diff
        return qpageview.diff.DiffRenderer()
    else:
        raise ValueError("unknown render type")
    r.paperColor = textformats.formatData('editor').baseColors['paper']
    return r


def loadPdf(filename):
    """Like qpageview.loadPdf(), but uses a preconfigured renderer."""
    return qpageview.loadPdf(filename, getRenderer("pdf"))


def loadSvgs(filenames):
    """Like qpageview.loadSvgs(), but uses a preconfigured renderer."""
    return qpageview.loadSvgs(filenames, getRenderer("svg"))


def loadImages(filenames):
    """Like qpageview.loadImages(), but uses a preconfigured renderer."""
    return qpageview.loadImages(filenames, getRenderer("image"))



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
QProgressBar subclasses.
"""


from PyQt6.QtCore import QTimeLine, QTimer
from PyQt6.QtWidgets import QProgressBar

class TimedProgressBar(QProgressBar):
    """A QProgressBar showing a certain time elapse."""

    def __init__(
        self,
        parent=None,
        hideWhileIdle=True,
        hidden=False,
        showFinished=3000
    ):
        super().__init__(parent, minimum=0, maximum=100)
        self._hideWhileIdle = hideWhileIdle
        self._hidden = hidden
        self._showFinished = showFinished
        self._timeline = QTimeLine(
            updateInterval=100, frameChanged=self.setValue
        )
        self._timeline.setFrameRange(0, 100)
        self._hideTimer = QTimer(
            timeout=self._done,
            singleShot=True,
            interval=showFinished
        )
        if not hidden and not hideWhileIdle:
            self.show()

    def setHidden(self, state):
        self._hidden = state
        self._done()

    def setHideWhileIdle(self, state):
        self._hideWhileIdle = state
        self._done()

    def setShowFinished(self, state):
        self._showFinished = state

    def start(self, total, elapsed=0.0):
        """Starts showing progress.

        total is the number of seconds (maybe float) the timeline will last,
        elapsed (defaulting to 0) is the value to start with.

        """
        self._hideTimer.stop()
        self._timeline.stop()
        self._timeline.setDuration(int(total * 1000))
        self._timeline.setCurrentTime(int(elapsed * 1000))
        self.setValue(self._timeline.currentFrame())
        self._timeline.resume()
        if self._hideWhileIdle and not self._hidden:
            self.show()

    def stop(self):
        """Ends the progress display.

        If showFinished is True (the default), 100% is shown for a few
        seconds and then the progress is reset.
        The progressbar is hidden if the hideOnTimeout attribute is True.

        """
        self._hideTimer.stop()
        self._timeline.stop()
        if self._showFinished and self._hideWhileIdle and not self._hidden:
            self.setValue(100)
            self._hideTimer.start()
        else:
            self._done()

    def _done(self):
        if self._hideWhileIdle and not self._hidden:
            self.hide()
        self.reset()



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Manages the progress bar in the status bar of ViewSpaces.
"""


from PyQt6.QtCore import Qt

import app
import job
import plugin
import metainfo
import widgets.progressbar

metainfo.define('buildtime', 0.0, float)


class ProgressBar(plugin.ViewSpacePlugin):
    """A Simple progress bar to show a Job is running."""

    def __init__(self, viewSpace):
        bar = self._bar = widgets.progressbar.TimedProgressBar(
            hideWhileIdle=True
        )
        viewSpace.status.layout().addWidget(bar, 0, Qt.AlignmentFlag.AlignCenter)
        bar.hide()
        viewSpace.viewChanged.connect(self.viewChanged)
        app.jobStarted.connect(self.jobStarted)
        app.jobFinished.connect(self.jobFinished)

    def viewChanged(self, view):
        self.showProgress(view.document())

    def showProgress(self, document):
        j = job.manager.job(document)
        if j and j.is_running():
            buildtime = metainfo.info(document).buildtime
            if not buildtime:
                # very arbitrary estimate...
                buildtime = 3.0 + document.blockCount() / 20
            self._bar.start(buildtime, j.elapsed_time())
            if job.attributes.get(j).hidden:
                self._bar.setEnabled(False)
                self._bar.setMaximumHeight(8)
                self._bar.setTextVisible(False)
            else:
                self._bar.setEnabled(True)
                self._bar.setMaximumHeight(14)
                self._bar.setTextVisible(True)
        else:
            self._bar.stop()

    def jobStarted(self, document, job):
        if document == self.viewSpace().document():
            self.showProgress(document)

    def jobFinished(self, document, j, success):
        if document == self.viewSpace().document():
            self._bar.setShowFinished(
                success and not job.attributes.get(j).hidden
            )
            self._bar.stop()
            if success:
                metainfo.info(document).buildtime = j.elapsed_time()


app.viewSpaceCreated.connect(ProgressBar.instance)



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Exception dialog for unhandled Python exceptions
(which are bugs in our program).
"""


import traceback

from PyQt6.QtCore import Qt
from PyQt6.QtGui import QTextCursor
from PyQt6.QtWidgets import (
    QDialog, QDialogButtonBox, QLabel, QTextBrowser, QVBoxLayout)

import app
import icons
import widgets
import bugreport


class ExceptionDialog(QDialog):
    """Single-use dialog displaying a Python exception."""
    def __init__(self, exctype, excvalue, exctb):
        super().__init__()

        # _tbshort is the exception line only (last line of the traceback)
        self._tbshort = ''.join(traceback.format_exception_only(exctype, excvalue))
        # _tbfull is the full traceback
        tbfull = traceback.format_exception(exctype, excvalue, exctb)
        self._tbfull = ''.join(tbfull)
        # _tblimited is the traceback, truncated to 5 frames max.
        # We do this because of size limits on what you can pass to GitHub
        # in a URL.
        tblimited = traceback.format_exception(exctype, excvalue, exctb, limit=-5)
        self._tblimited = ''.join(tblimited)

        self._ext_maintainer = app.extensions().is_extension_exception(exctb)

        layout = QVBoxLayout()
        self.setLayout(layout)

        self.errorLabel = QLabel()
        layout.addWidget(self.errorLabel)
        textview = QTextBrowser()
        layout.addWidget(textview)
        textview.setText(self._tbfull)
        textview.moveCursor(QTextCursor.MoveOperation.End)

        layout.addWidget(widgets.Separator())

        self.infoLabel = QLabel()
        self.infoLabel.setWordWrap(True)
        self.infoLabel.setOpenExternalLinks(True)
        layout.addWidget(self.infoLabel)

        b = self.buttons = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        b.button(QDialogButtonBox.StandardButton.Ok).setIcon(icons.get("tools-report-bug"))
        layout.addWidget(b)

        b.accepted.connect(self.accept)
        b.rejected.connect(self.reject)
        self.resize(600,300)
        app.translateUI(self)
        self.exec()

    def translateUI(self):
        if self._ext_maintainer:
            extension = self._ext_maintainer[0]
            text = _("An internal error has occurred in extension '{name}':").format(name=extension)
            title = _("Extension Error")
            info = _("Please use the button below to send a bug report to the "
                     "maintainer of this extension by email.")
        else:
            text = _("An internal error has occurred:")
            title = _("Internal Error")
            info = _("We would highly appreciate if you used the button below "
                     "to open a new issue on GitHub, the development platform "
                     "used by the Frescobaldi project, to let developers "
                     "know about this problem. If you have general questions, "
                     "you can also send them to the <a href=\"https://groups.google.com/g/frescobaldi\">"
                     "Frescobaldi user group</a>.")
        self.setWindowTitle(app.caption(title))
        self.errorLabel.setText(text)
        self.infoLabel.setText(info)
        self.buttons.button(QDialogButtonBox.StandardButton.Ok).setText(_("Send Bug Report..."))

    def done(self, result):
        if result:
            self.reportBug()
        super().done(result)

    def reportBug(self):
        if self._ext_maintainer:
            # For now extensions don't have a way to use a custom bug reporting
            # method, we use email.
            extension = self._ext_maintainer[0]
            rcpt = self._ext_maintainer[1]
            ext_intro = '\n{}\n\n'.format(
                _("An error occurred in extension '{name}':").format(name=extension))
            ext_header = f' [{self._ext_maintainer[0]}]'
            bugreport.email(
                self._tbshort + ext_header,
                ext_intro + self._tbfull + '\n'
                + _("Please describe below what you were doing:"),
                recipient=rcpt)
        else:
            body = (self._tblimited + '\n'
                    + _("Please describe below what you were doing:"))
            bugreport.new_github_issue(title=self._tbshort, body=body)



# process.py -- A very simple wrapper around QProcess
#
# Copyright (c) 2012 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
A (multithreaded) Job queue
"""

from enum import Enum
import collections
import time

from PyQt6.QtCore import QObject

import app
import signals


class RunnerBusyException(Exception):
    """Raised when a Runner is asked to start a job (without the force=True
    keyword argument) while having already a running job."""
    pass


class Runner(QObject):
    """A Runner in the stack of a JobQueue.

    Responsible for executing a single job.Job instance.
    Receives a job through the start command from the queue, reports
    back about completion of the job. Any further actions are
    managed by the queue.

    References are kept to the queue and the job as well as the index
    in the queue.
    """

    def __init__(self, queue, index):
        super().__init__()
        self._queue = queue
        self._index = index
        self._job = None
        self._completed = 0

    def abort(self):
        """Aborts a running job if any."""
        if self._job and self._job.is_running():
            self._job.abort()

    def completed(self):
        """Return the number of jobs completed during the Runner's lifetime."""
        return self._completed

    def index(self):
        """Return the index of the Runner in the JobQueue."""
        return self._index

    def is_running(self):
        return self._job and self._job.is_running()

    def job(self):
        return self._job

    def job_done(self):
        """Count job, notify queue, remove reference to Job object."""
        self._completed += 1
        job = self._job
        self._job = None
        self._queue.job_completed(self, job)

    def start(self, j, force=False):
        """Start a given job.
        If currently a job is running either abort that
        or raise an exception."""
        if self._job and self.is_running():
            if force:
                self.abort()
            else:
                raise RunnerBusyException(
                    _("Job is already running. Wait for completion."))
        self._job = j
        j.set_runner(self)
        j.done.connect(self.job_done)
        j.start()


class AbstractQueue(QObject):
    """Common interface for the different queue types used in the JobQueue.
    The various queue classes are very lightweight wrappers around the
    corresponding concepts and base objects, with the only reason to
    provide a transparent interface for used in JobQueue."""

    def __init__(self):
        super().__init__()

    def clear(self):
        """Remove all entries from the queue."""
        raise NotImplementedError

    def empty(self):
        """Return True if there are no queued items."""
        return self.length() == 0

    def length(self):
        """Return the length of the queue. Only has to be overridden
        when the data structure doesn't support len()."""
        return len(self._queue)

    def push(self, j):
        """Add a job to the queue."""
        raise NotImplementedError

    def pop(self):
        """Remove and return the next job."""
        raise NotImplementedError


class AbstractStackQueue(AbstractQueue):
    """Common ancestor for LIFO and FIFO queues"""

    def __init__(self):
        super().__init__()
        self._queue = collections.deque()

    def clear(self):
        """Remove all entries from the queue."""
        self._queue.clear()

    def pop(self):
        return self._queue.pop()


class Queue(AbstractStackQueue):
    """First-in-first-out queue (default operation)."""

    def push(self, j):
        self._queue.appendleft(j)


class Stack(AbstractStackQueue):
    """Last-in-first-out queue, or stack."""

    def push(self, j):
        self._queue.append(j)


class PriorityQueue(AbstractQueue):
    """Priority queue, always popping the job with the highest priority.

    Uses Job's priority() property (which defaults to 1) and a transparent
    insert count to determine order of popping jobs. If jobs have the same
    priority they will be served first-in-first-out."""

    def __init__(self):
        super().__init__()
        self._queue = []
        self._insert_count = 0

    def clear(self):
        """Remove all entries from the queue."""
        self._queue = []

    def push(self, j):
        """Add a job to the queue. retrieve the priority from the job,
        add an autoincrement value for comparing jobs with identical
        priority."""
        from heapq import heappush
        heappush(self._queue, (j.priority(), self._insert_count, j))
        self._insert_count += 1

    def pop(self):
        """Return the correct part of the tuplet
        (1st: priority, 2nd: insert order)."""
        from heapq import heappop
        return heappop(self._queue)[2]


class JobQueueException(Exception):
    """Abstract base exception for JobQueue related exceptions."""
    pass


class JobQueueStateException(JobQueueException):
    """Raised when an operation is not allowed in the current
    state of the queue."""
    pass


class QueueStatus(Enum):
    INACTIVE = 0
    STARTED = 1
    PAUSED = 2
    EMPTY = 3
    IDLE = 4
    FINISHED = 5
    ABORTED = 6


class QueueMode(Enum):
    """Running mode of the JobQueue.
    CONTINUOUS means that the queue can be idle, waiting for new jobs,
    SINGLE means that it will be considered finished when running empty."""
    CONTINUOUS = 0
    SINGLE = 1


class JobQueue(QObject):
    """A (multiThreaded) Job Queue.

    Manages a given number of "runners" which can process one job
    at a time each.

    An arbitrary number of job.Job() instances (or any descendants)
    can be added to the queue, which are distributed in parallel
    to the runners.

    A special case is a queue with only one runner. This can be used
    to ensure that asynchronous jobs can run in sequence, either to
    ensure that only one such job runs at a time (e.g. to force
    long-running stuff to the background) or to ensure subsequent
    jobs can use the results of earlier ones.

    The queue supports pause(), resume() and abort() operations.
    It iterates through the states
    QueueStatus.INACTIVE
                STARTED
                PAUSED (remaining jobs won't be started)
                EMPTY (no *new* jobs are queued)
                IDLE (no new jobs available, all jobs completed)
                FINISHED
                ABORTED

    A Queue can work in two modes: QueueMode.CONTINUOUS (default) and
    QueueMode.SINGLE. In single mode the queue is considered finished once
    there are no jobs available while a continuous queue switches to
    IDLE in that case. A CONTINUOUS queue is always considered to be
    active (idle when empty) and doesn't have to be explicitly started.

    If a 'capacity' is passed to the queue it has the notion of "full",
    otherwise an unlimited number of jobs can be enqueued.

    By default an internal FIFO (First in, first out) Queue is used
    as the underlying data structure, but Stack and PriorityQueue are
    available through the keyword command as well.
    """

    started = signals.Signal()
    paused = signals.Signal()
    resumed = signals.Signal()
    emptied = signals.Signal()  # emitted when last job is popped
    idle = signals.Signal()  # emitted when waiting for new jobs
                             # after the last job has been completed
    finished = signals.Signal()
    aborted = signals.Signal()
    # The following three signals emit the corresponding job as argument.
    job_added = signals.Signal()
    job_done = signals.Signal()  # emitted when a job has been completed.
               # When this is emitted, the queue's state has been
               # updated (other than with the *job's* signal)
    job_started = signals.Signal()

    def __init__(self,
                 queue_class=Queue,
                 queue_mode=QueueMode.CONTINUOUS,
                 num_runners=1,
                 tick_interval=1000,
                 capacity=None):
        super().__init__()
        self._state = QueueStatus.INACTIVE
        self._queue_mode = queue_mode
        self._starttime = None
        self._endtime = None
        self._completed = 0
        self._queue = queue_class()
        self._capacity = capacity
        self._runners = [Runner(self, i) for i in range(num_runners)]

        if queue_mode == QueueMode.CONTINUOUS:
            self.start()

    def abort(self, force=True):
        """Abort the execution of the queued jobs.
        If force=True running jobs are aborted, otherwise
        only the queue is cleared, allowing running jobs to finish."""
        if self.state() in [
            QueueStatus.FINISHED,
            QueueStatus.ABORTED
        ]:
            raise JobQueueStateException(
                _("Inactive Job Queue can't be aborted")
            )
        self.set_state(QueueStatus.ABORTED)
        self.set_queue_mode(QueueMode.SINGLE)
        self._queue.clear()
        if force:
            for runner in self._runners:
                if runner:
                    # ignore runners that have already been set to None
                    runner.abort()
        self.aborted.emit()

    def add_job(self, job):
        """Enqueue a new job to the queue.

        Some checks are made to determine the validity of adding a new job.
        If the queue hasn't started yet or is in pause the job is simply
        pushed to the queue, otherwise it will be determined whether an
        idle runner is available to start with the job immediately.
        """
        if self.full():
            raise IndexError(_("Job Queue full"))
        if self.state() in [QueueStatus.FINISHED, QueueStatus.ABORTED]:
            raise JobQueueStateException(
                _("Can't add job to finished/aborted queue.")
            )
        elif self.state() in [QueueStatus.INACTIVE, QueueStatus.PAUSED]:
            self._queue.push(job)
            self.job_added.emit(job)
        else:
            runner = self.idle_runner()
            if runner:
                self.job_added.emit(job)
                runner.start(job)
                self.job_started.emit(job)
            else:
                self._queue.push(job)
                self.job_added.emit(job)
            self.set_state(
                QueueStatus.EMPTY if self._queue.empty()
                else QueueStatus.STARTED)

    def completed(self, runner=-1):
        """Return the number of completed jobs,
        either for a given runner or the sum of all runners."""
        if runner >= 0:
            return self._runners[runner].completed()
        else:
            result = 0
            for i in range(len(self._runners)):
                result += self._runners[i].completed()
            return result

    def full(self):
        """Returns True if a maximum capacity is set and used."""
        if not self._capacity:
            return False
        return self._queue.length() == self._capacity

    def is_idle(self):
        """Returns True if all Runners are idle."""
        for runner in self._runners:
            if runner.is_running():
                return False
        return True

    def is_running(self):
        """Return True if the queue is 'live'."""
        return self.state() not in ([
            QueueStatus.INACTIVE,
            QueueStatus.FINISHED,
            QueueStatus.ABORTED
        ])

    def idle_runner(self):
        """Returns the first idle Runner object,
        or None if all are busy."""
        for runner in self._runners:
            if not runner.is_running():
                return runner
        return None

    def job_completed(self, runner, job):
        """Called by a runner once its job has completed.

        Manage behaviour at that point, depending on the
        queue's state and mode.
        """
        if self.state() == QueueStatus.STARTED:
            runner.start(self.pop())
        elif self.state() == QueueStatus.PAUSED:
            # If a SINGLE queue completes the last job while in PAUSE mode
            # it can be considered finished.
            if (
                self._queue.empty()
                and self.is_idle()
                and self.queue_mode() == QueueMode.SINGLE
            ):
                self.queue_finished()
        elif self.is_idle():
            # last runner has completed its job and queue is empty.
            # Either finish queue or set to IDLE.
            if self.queue_mode() == QueueMode.SINGLE:
                self.queue_finished()
            else:
                self.set_state(QueueStatus.IDLE)
                self.idle.emit()
        self.job_done.emit(job)

    def pause(self):
        """Pauses the execution of the queue.
        Running jobs are allowed to finish, but no new jobs will be started.
        (The actual behaviour is implemented in job_completed().)"""
        if self.state() in [
            QueueStatus.INACTIVE,
            QueueStatus.FINISHED,
            QueueStatus.ABORTED
        ]:
            raise JobQueueStateException(
                _("Non-running Job Queue can't be paused.")
            )
        self.set_state(QueueStatus.PAUSED)
        self.paused.emit()

    def pop(self):
        """Return and remove the next Job.
        Raises Exception if empty."""
        if self.state() == QueueStatus.EMPTY:
            raise IndexError("Job Queue is empty.")
        if self.state() != QueueStatus.STARTED:
            raise JobQueueStateException(
                _("Can't pop job from non-started Job Queue")
            )
        j = self._queue.pop()
        if self._queue.empty():
            self.set_state(QueueStatus.EMPTY)
            self.emptied.emit()
        return j

    def queue_finished(self):
        """Called when the last job has been completed and the queue
        is in SINGLE mode."""
        if self.state() != QueueStatus.ABORTED:
            self.set_state(QueueStatus.FINISHED)
        self.finished.emit()

    def queue_mode(self):
        return self._queue_mode

    def resume(self):
        """Resume the queue from PAUSED state by trying to start
        all runners."""
        if not self.state() == QueueStatus.PAUSED:
            raise JobQueueStateException(
                _("Job Queue not paused, can't resume.")
            )
        self._start()
        self.resumed.emit()

    def set_idle(self):
        """Set status to IDLE if all runners are in idle mode."""
        for runner in self._runners:
            if runner.is_running():
                return
        self.set_state(QueueStatus.IDLE)
        self.idle.emit()

    def set_queue_mode(self, mode):
        self._queue_mode = mode

    def size(self):
        """Return the number of unstarted jobs."""
        return self._queue.length()

    def _start(self):
        """Set the state to started and ask all runners to start."""
        if self.state() in [QueueStatus.FINISHED, QueueStatus.ABORTED]:
            raise JobQueueStateException(
                _("Can't (re)start a finished/aborted Job Queue.")
            )
        elif self.state() == QueueStatus.STARTED:
            raise JobQueueStateException(_("Queue already started."))
        if self.state() in [
            QueueStatus.INACTIVE,
            QueueStatus.PAUSED
        ]:
            self.set_state(QueueStatus.STARTED)

        if (
            self.state() == QueueStatus.EMPTY
            and self.queue_mode() == QueueMode.SINGLE
        ):
            raise IndexError(_("Can't start SINGLE-mode empty queue"))
        if self._queue.empty():
            self.set_state(QueueStatus.IDLE)
        else:
            self.set_state(QueueStatus.STARTED)
            for runner in self._runners:
                if self._queue.empty():
                    break
                if not runner.is_running():
                    runner.start(self.pop())

    def start(self):
        """Start processing of the queue."""
        if self.is_running():
            raise JobQueueStateException(
                _("Can't 'start' an active Job Queue."))
        self._starttime = time.time()
        self._start()
        self.started.emit()

    def set_state(self, state):
        self._state = state

    def state(self):
        return self._state


class GlobalJobQueue(QObject):
    """The application-wide Job Queue that dispatches jobs to runners
    and subordinate queues.
    """

    def __init__(self):
        self.load_settings()
        self._crawler = JobQueue()
        self._engraver = JobQueue()
        self._generic = JobQueue()
        self._queues = {
            'crawl': self._crawler,
            'engrave': self._engraver,
            'generic': self._generic
        }
        app.settingsChanged.connect(self.settings_changed)
        app.aboutToQuit.connect(self.about_to_quit)

    def about_to_quit(self):
        # TODO:
        # - is any queue active?
        # - should jobs be aborted?
        # - or only the queues be emptied?
        # - _crawer can always be aborted immediately
        pass

    def add_job(self, j, target='engrave'):
        """Add a job to the specified job queue."""
        target_queue = self._queues.get(target, None)
        if not target_queue:
            raise ValueError(_("Invalid job queue target: {name}").format(name=target))
        target_queue.add_job(j)

    def load_settings(self):
        # TODO: Load settings and create the JobQueues accordingly
        pass

    def settings_changed(self):
        # TODO: If multicore-related settings have changed update the queues
        pass



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
A widget and dialog to show an output preview of a LilyPond document.
"""


import os

from PyQt6.QtCore import (
    QSize,
    Qt
)
from PyQt6.QtWidgets import (QComboBox, QDialog, QDialogButtonBox, QHBoxLayout,
                             QLabel, QStackedLayout, QVBoxLayout, QWidget)

import app
import icons
import job.lilypond
import log
import qutil
import pagedview
import widgets.progressbar


class MusicPreviewWidget(QWidget):
    def __init__(
        self,
        parent=None,
        showProgress=True,
        showWaiting=False,
        progressHidden=False,
        progressHiddenWhileIdle=True,
        progressShowFinished=3000,
        showLog=True
    ):
        super().__init__(parent)
        self._lastbuildtime = 10.0
        self._running = None
        self._current = None

        self._showLog = showLog
        if showLog:
            self._log = log.Log()
        self._showProgress = showProgress

        self._chooserLabel = QLabel()
        self._chooser = QComboBox(self, activated=self.selectDocument)
        self._view = pagedview.PagedView()
        self._view.setMagnifier(pagedview.Magnifier())

        self._showWaiting = showWaiting
        if showWaiting:
            from widgets.waitingoverlay import Overlay
            self._waiting = Overlay(self._view)
            self._waiting.hide()

        self._stack = QStackedLayout()
        self._top = QWidget()

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        layout.addWidget(self._top)
        layout.addLayout(self._stack)
        if self._showProgress:
            self._progress = widgets.progressbar.TimedProgressBar(
                parent=self,
                hidden=progressHidden,
                hideWhileIdle=progressHiddenWhileIdle,
                showFinished=progressShowFinished
            )
            layout.addWidget(self._progress)

        top = QHBoxLayout()
        top.setContentsMargins(0, 0, 0, 0)
        top.setSpacing(2)
        self._top.setLayout(top)
        top.addWidget(self._chooserLabel)
        top.addWidget(self._chooser)
        top.addStretch(1)

        if showLog:
            self._stack.addWidget(self._log)
        self._stack.addWidget(self._view)

        self._top.hide()
        app.aboutToQuit.connect(self.cleanup)
        app.translateUI(self)

    def translateUI(self):
        self._chooserLabel.setText(_("Document:"))

    def abort_running(self):
        """Ensures no running job is left behind.
        This *can* be called by the using dialog/widget."""
        j = self._running
        if j and j.is_running():
            if self._showLog:
                self._log.disconnectJob(j)
            j.done.disconnect(self._done)
            j.abort()

    def preview(
        self, text, title=None, base_dir=None,
        temp_dir='', cached=False
    ):
        """Runs LilyPond on the given text and shows the resulting PDF."""
        self.abort_running()
        if cached:
            self._running = j = job.lilypond.CachedPreviewJob(
                text,
                target_dir=temp_dir,
                base_dir=base_dir,
                title=title
            )
            if not self._running.needs_compilation():
                self._done(None)
                return
        else:
            self._running = j = job.lilypond.VolatileTextJob(
                text,
                title=title
            )
        j.done.connect(self._done)
        if self._showLog:
            self._log.clear()
            self._log.connectJob(j)
            self._stack.setCurrentWidget(self._log)
        if self._showProgress:
            j.started.connect(
                lambda: self._progress.start(self._lastbuildtime)
            )
            self._progress.start(self._lastbuildtime)
        if self._showWaiting:
            self._waiting.start()
        app.job_queue().add_job(j, 'generic')

    def _done(self, success):
        # TODO: Handle failed compilation (= no file to show)
        if self._showProgress:
            self._progress.stop()
        if self._showWaiting:
            self._waiting.stop()
        pdfs = self._running.resultfiles()
        self.setDocuments(pdfs)
        if not pdfs and self._showLog:
            self._stack.setCurrentWidget(self._log)
            return
        self._lastbuildtime = self._running.elapsed_time()
        self._stack.setCurrentWidget(self._view)
        if self._current:
            self._current.cleanup()
        self._current = self._running  # keep the tempdir
        self._running = None

    def setDocuments(self, pdfs):
        """Loads the given PDF path names in the UI."""
        self._documents = [pagedview.loadPdf(name) for name in pdfs]
        self._chooser.clear()
        self._chooser.addItems([os.path.basename(d.filename()) for d in self._documents])
        self._top.setVisible(len(self._documents) > 1)
        if pdfs:
            self._chooser.setCurrentIndex(0)
            self.selectDocument(0)
        else:
            self._view.clear()

    def selectDocument(self, index):
        self._view.setDocument(self._documents[index])

    def cleanup(self):
        if self._running:
            self._running.abort()
            self._running.cleanup()
            self._running = None
        if self._current:
            self._current.cleanup()
            self._current = None
        if self._showLog:
            self._stack.setCurrentWidget(self._log)
        self._top.hide()
        self._view.clear()

    def print_(self):
        """Prints the currently displayed document."""
        if self._view.document():
            self._view.print()


class MusicPreviewDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        layout = QVBoxLayout()
        self.setLayout(layout)
        self._widget = MusicPreviewWidget()
        layout.addWidget(self._widget)
        layout.addWidget(widgets.Separator())
        b = QDialogButtonBox()
        layout.addWidget(b)
        b.addButton(QDialogButtonBox.StandardButton.Close)
        b.rejected.connect(self.accept)
        self._printButton = b.addButton('', QDialogButtonBox.ButtonRole.ActionRole)
        self._printButton.setIcon(icons.get("document-print"))
        self._printButton.clicked.connect(self._widget.print_)
        self._printButton.hide()
        qutil.saveDialogSize(self, "musicpreview/dialog/size", QSize(500, 350))
        app.translateUI(self)

    def translateUI(self):
        self._printButton.setText(_("&Print"))
        self.setWindowTitle(app.caption(_("Music Preview")))

    def preview(self, text, title=None):
        self._widget.preview(text, title)

    def cleanup(self):
        self._widget.cleanup()

    def setEnablePrintButton(self, enable):
        """Enables or disables the print button."""
        self._printButton.setVisible(enable)



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.


"""
Download music fonts from the openlilylib-resources Github repository
"""

CATALOGUE_URL = (
    "https://raw.githubusercontent.com/openlilylib-resources/"
    "lilypond-notation-fonts/master/catalogue.json"
)

# import urllib.request
# import json
# try:
#     response = urllib.request.urlopen(CATALOGUE_URL)
#     data = repsonse.read()
#     catalogue = json.loads(data.decode('utf-8'))
# ...

# TODO:
# - only works when there's a music font repository set
# - download catalogue
# - compare available fonts with installed fonts (incl. version check)
# - open a dialog with a list of check boxes for all available fonts
#   (fonts that don't need to be updated are initially unchecked,
#    but it's possible to overwrite them)
#
# - download to temp dir first,
#   and when that's successful copy/overwrite to the local repository
# - install any newly downloaded font (in *this* case not skipping
#   but overwriting existing fonts)



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
A message box suggesting a restart.

Some operations make it necessary to restart Frescobaldi,
either for simply taking effect or to avoid possible instabilities.
These include switching the Git branch or removing extensions.
"""

from PyQt6.QtWidgets import QMessageBox

import app

def suggest_restart(operation):
    """Show a message box informing about the need to
    restart the application. the 'operation' argument
    is used to customize the message."""

    msgBox = QMessageBox()
    msgBox.setIcon(QMessageBox.Icon.Warning)
    msgBox.setText(_("Restart Required"))

    # Translate first, then use format
    translated_template = _(
        "An operation requires a restart of Frescobaldi:\n\n"
        "   {}\n\n"
        "Some new functionality may not be available before, "
        "or the application may even become unstable.\n\n"
        "Do you want to restart now?\n"
        "You can also save open files first and restart manually."
    )
    formatted_message = translated_template.format(operation)
    msgBox.setInformativeText(formatted_message)

    msgBox.setStandardButtons(QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel)
    if msgBox.exec() == QMessageBox.StandardButton.Ok:
        app.activeWindow().restart()



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Generic import dialog. Presuppose a child instance for the specific import.
"""


import os
import platform

from PyQt6.QtCore import Qt

from PyQt6.QtWidgets import (QCheckBox, QDialog, QDialogButtonBox,
    QGridLayout, QLabel, QTabWidget, QTextEdit, QWidget)

import app
import lilychooser
import userguide
import util
import widgets
import job


class ToLyDialog(QDialog):
    """Dialog base class for all file import jobs.

    Basically the whole dialog is identical for all file types, the only
    part of the dialog that is specific is the widget with *input* options.
    These are created in a subclass's __init__ function and added to the
    self.impChecks list. Only then super() is called.
    """

    def __init__(self,
                 parent=None,
                 job_class=job.Job,
                 imp_prgm='',
                 input=None,
                 userg=''):
        super().__init__(parent)
        self._info = None
        self._imp_prgm = imp_prgm
        self._userg = userg
        self._input = input
        self._path = None
        self._job_class = job_class
        self._job = None

        self.addAction(parent.actionCollection.help_whatsthis)
        self.setWindowModality(Qt.WindowModality.WindowModal)
        mainLayout = QGridLayout()
        self.setLayout(mainLayout)

        tabs = QTabWidget()

        import_tab = QWidget()
        post_tab = QWidget()

        itabLayout = QGridLayout(import_tab)
        ptabLayout = QGridLayout(post_tab)

        tabs.addTab(import_tab, self._imp_prgm)
        tabs.addTab(post_tab, _("After Import"))

        self.formatCheck = QCheckBox()
        self.trimDurCheck = QCheckBox()
        self.removeScalesCheck = QCheckBox()
        self.runEngraverCheck = QCheckBox()

        self.postChecks = [self.formatCheck,
                           self.trimDurCheck,
                           self.removeScalesCheck,
                           self.runEngraverCheck]

        self.versionLabel = QLabel()
        self.lilyChooser = lilychooser.LilyChooser(toolcommand=self._imp_prgm)

        self.formatCheck.setObjectName("reformat")
        self.trimDurCheck.setObjectName("trim-durations")
        self.removeScalesCheck.setObjectName("remove-scaling")
        self.runEngraverCheck.setObjectName("engrave-directly")

        self.buttons = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        userguide.addButton(self.buttons, self._userg)

        row = 0
        for r, w in enumerate(self.impChecks):
            row += r
            itabLayout.addWidget(w, row, 0, 1, 2)
        row += 1
        for r, w in enumerate(self.impExtra):
            row += r
            itabLayout.addWidget(w, row, 0, 1, 2)

        itabLayout.addWidget(widgets.Separator(), row + 1, 0, 1, 2)
        itabLayout.addWidget(self.versionLabel, row + 2, 0, 1, 0)
        itabLayout.addWidget(self.lilyChooser, row + 3, 0, 1, 2)
        itabLayout.setRowStretch(row + 4, 10)

        ptabLayout.addWidget(self.formatCheck, 0, 0, 1, 2)
        ptabLayout.addWidget(self.trimDurCheck, 1, 0, 1, 2)
        ptabLayout.addWidget(self.removeScalesCheck, 2, 0, 1, 2)
        ptabLayout.addWidget(self.runEngraverCheck, 3, 0, 1, 2)
        ptabLayout.setRowStretch(4, 6)

        mainLayout.addWidget(tabs, 0, 0, 6, 2)
        mainLayout.addWidget(self.buttons, 7, 0, 1, 2)

        self.buttons.accepted.connect(self.about_to_accept)
        self.buttons.rejected.connect(self.reject)

        self.lilyChooser.currentIndexChanged.connect(self.slot_lilypond_version_changed)
        self.slot_lilypond_version_changed()

    def translateUI(self):
        self.versionLabel.setText(_("LilyPond version:"))
        self.formatCheck.setText(_("Reformat source"))
        self.trimDurCheck.setText(_("Trim durations (Make implicit per line)"))
        self.removeScalesCheck.setText(_("Remove fraction duration scaling"))
        self.runEngraverCheck.setText(_("Engrave directly"))

    def about_to_accept(self):
        """Configure the job and close the dialog."""
        self.configure_job()
        self.accept()

    def configure_job(self):
        """Create and configure the job to be run.
        Has to be completed by the subclasses."""
        output = os.path.splitext(
            os.path.join(util.tempdir(), os.path.basename(self._input))
            )[0] + '.ly'
        self._job = j = self._job_class(
            command=self._info.toolcommand(self._imp_prgm),
            input=self._input,
            output=f'--output={output}',
            directory=os.path.dirname(self._input),
            encoding='utf-8')
        j._output_file = output
        if platform.system() == "Darwin":
            import macos
            if macos.inside_app_bundle():
                j.environment['PYTHONPATH'] = None
                j.environment['PYTHONHOME'] = None

    def get_post_settings(self):
        """Returns settings in the post import tab."""
        post = []
        for p in self.postChecks:
            post.append(p.isChecked())
        return post

    def job(self):
        """Return the current job."""
        if not self._job:
            self.configure_job()
        return self._job

    def set_input(self, path):
        """Set the full path to the input document."""
        self._input = path

    def slot_lilypond_version_changed(self):
        self._info = self.lilyChooser.lilyPondInfo()

    def loadSettings(self):
        """Get users previous settings."""
        post_default = [True, False, False, True]
        for i, d in zip(self.impChecks, self.imp_default):
            i.setChecked(self.settings.value(i.objectName(), d, bool))
        for p, f in zip(self.postChecks, post_default):
            p.setChecked(self.settings.value(p.objectName(), f, bool))

    def saveSettings(self):
        """Save users last settings."""
        for i in self.impChecks:
            self.settings.setValue(i.objectName(), i.isChecked())
        for p in self.postChecks:
            self.settings.setValue(p.objectName(), p.isChecked())



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Interface for extensions to save and load settings from extension files
"""

import os
import sys

from PyQt6.QtCore import Qt, QObject, QSettings
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import QMenu

import app

class ExtensionSettings(QObject):
    """
    Manage an extension's settings, serves as an abstraction to QSettings
    to make it easier to handle in an extension and to make it harder
    to create damage.
    Allowed settings and their default values (and implicit types) are defined
    in the Extension class with the _settings_config class variable.
    Extension code should never work with QSettings themselves but interact
    with the extension.settings() object. Settings are loaded once upon startup
    and then maintained in the object. Settings are retrieved thorugh
    settings().get() and changed with settings().set(). When setting a value
    both the settings object is updated and the setting written to disk
    immediately.
    Settings are managed in a 'settings.conf' file in the extension directory.
    """
    def __init__(self, extension):
        super().__init__(extension)
        self._extension = extension
        self._defaults = extension._settings_config
        self._data = dict(self._defaults)
        self._load_settings()

    def extension(self):
        return self._extension

    def get(self, key):
        """Retrieve a setting.
        Raises a KeyError if non-existent key is requested."""
        if key not in self._defaults.keys():
            raise KeyError(
                _("Trying to retrieve unknown setting '{setting}' "
                  "in extension '{extension}'").format(
                    setting=key,
                    extension=self.extension().name()))
        return self._data[key]

    def _load_settings(self):
        """Load settings from disk, is done once upon creation."""
        s = QSettings(self.settings_file(), QSettings.Format.IniFormat)
        for key in self._defaults:
            default = self._defaults[key]
            self._data[key] = s.value(key, default, type(default))

    def set(self, key, value):
        """Set a setting and save it to disk.
        Check type before and raise an exception if necessary."""
        extension_name = (self.extension().display_name()
            or self.extension().name())
        if key not in self._defaults.keys():
            raise KeyError(
                _("Trying to store unknown setting '{setting}' "
                  "in extension '{extension}'").format(
                    setting=key,
                    extension=extension_name))
        try:
            old_value = self._data[key]
            value = type(self._defaults[key])(value)
            if value == old_value:
                return
            self._data[key] = value
            QSettings(self.settings_file(), QSettings.Format.IniFormat).setValue(key, value)
            self.extension().settings_changed(key, old_value, value)
        except ValueError:
            # type conversion failed
            raise ValueError(
                _("Trying to store setting '{setting}' in extension "
                  "'{extension}' with type '{type}' instead of '{type_is}'").format(
                  setting=key,
                  extension=extension_name,
                  type = type(self._defaults[key]),
                  type_is = type(value)))

    def settings_file(self):
        """Return path to 'settings.conf' in the extension directory."""
        extension = self.extension()
        return os.path.join(
            extension.parent().root_directory(),
            extension.name(),
            'settings.conf')



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

import os
import re

from PyQt6.QtCore import (
    QObject,
    QRegularExpression,
    QSettings,
    QSortFilterProxyModel,
    Qt,
)
from PyQt6.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QLabel,
    QLineEdit,
    QMenu,
    QTreeView,
    QVBoxLayout,
    QWidget,
)
from PyQt6.QtGui import (
    QAction,
    QFont,
    QFontDatabase,
    QStandardItem,
    QStandardItemModel,
)

import app
import job
import signals


# List of notation fonts currently installed.
_installed_notation_fonts = []


class TextFontsWidget(QWidget):
    """Display installed text fonts available for a given LilyPond version."""
    # Store the filter expression over the object's lifetime
    filter_re = ''

    def __init__(self, dialog):
        super().__init__(dialog)
        available_fonts = dialog.available_fonts
        self.lilypond_info = available_fonts.lilypond_info
        self.fonts = available_fonts.text_fonts()

        self.status_label = QLabel(wordWrap=True)
        self.tree_view = tv = QTreeView()
        self.filter_edit = QLineEdit(clearButtonEnabled=True)
        tv.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        tv.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        tv.customContextMenuRequested.connect(self.show_context_menu)

        layout = QVBoxLayout(self)
        layout.addWidget(self.status_label)
        layout.addWidget(self.tree_view)
        layout.addWidget(self.filter_edit)
        self.setLayout(layout)

        self.tree_view.setModel(self.fonts.model().proxy())
        self.filter_edit.textChanged.connect(self.update_filter)
        self.loadSettings()
        dialog.finished.connect(self.saveSettings)
        app.translateUI(self)

        if self.fonts.is_loaded():
            self.populate()
        else:
            self.display_waiting()
            self.fonts.loaded.connect(self.populate)

    def translateUI(self):
        self.filter_edit.setPlaceholderText(_(
            "Filter results (type any part of the font family name. "
            "Regular Expressions supported.)"
        ))

    def loadSettings(self):
        s = QSettings()
        s.beginGroup('document-fonts-dialog')
        self.tree_view.setColumnWidth(0, int(s.value('col-width', 200)))
        self.filter = QRegularExpression('')
        self.filter.setPatternOptions(
            QRegularExpression.PatternOption.CaseInsensitiveOption)

    def saveSettings(self):
        # Text font tab
        s = QSettings()
        s.beginGroup('document-fonts-dialog')
        s.setValue('col-width', self.tree_view.columnWidth(0))

    def display_count(self):
        self.status_label.setText(
            _("{count} font families detected by {version}").format(
                count=self.tree_view.model().rowCount(),
                version=self.lilypond_info.prettyName()))

    def display_waiting(self):
        self.status_label.setText(_("Running LilyPond to list fonts ..."))

    def load_font_tree_column_width(self):
        """Load column widths for fontTreeView,
        factored out because it has to be done upon reload too."""
        s = QSettings()
        s.beginGroup('document-fonts-dialog')
        self.tree_view.setColumnWidth(0, int(s.value('col-width', 200)))

    def populate(self):
        self.load_font_tree_column_width()
        self.display_count()
        self.refresh_filter_edit()
        self.filter_edit.setFocus()
        QApplication.restoreOverrideCursor()

    def refresh_filter_edit(self):
        self.filter_edit.setText(TextFontsWidget.filter_re)

    def set_text_font(self):
        """
        Determine the currently selected font name
        and set it to the given font family (roman, sans, typewriter).
        Triggers recompilation of the example (if the font type family
        is selected in the font command tab and the example isn't cached).
        """
        family = self.sender().family
        indexes = self.tree_view.selectionModel().selectedIndexes()
        font_name = (
            indexes[0].data()
            # Row with font name has been selected
            if len(indexes) == 1 or not indexes[1].data()
            # Row with font weight has been selected (second col has sample)
            else indexes[0].parent().data()
        )
        self.window().select_font(family, font_name)
        self.window().invalidate_command()

    def show_context_menu(self, point):
        """Show a context menu to set text font families."""
        cm = QMenu(self)
        actions = {}
        for family in ['Roman', 'Sans', 'Typewriter']:
            f_key = family.lower()
            actions[family] = ac = QAction()
            ac.setText(
                _('Set as {family} (current: {current})').format(
                    family=family,
                    current=self.window().selected_font(f_key)
                ))
            ac.family = f_key
            ac.triggered.connect(self.set_text_font)
            cm.addAction(actions[family])
        cm.exec(self.tree_view.mapToGlobal(point))
        cm.deleteLater()

    def tree_model(self):
        return self.tree_view.model()

    def update_filter(self):
        """Filter font results"""
        TextFontsWidget.filter_re = re = self.filter_edit.text()
        self.filter.setPattern(re)
        self.fonts.model().proxy().setFilterRegularExpression(self.filter)


class MiscFontsInfoWidget(QWidget):
    """Display miscellaneous info about Fontconfig context."""
    def __init__(self, available_fonts, parent=None):
        super().__init__(parent)
        self.tree_view = tv = QTreeView(self)
        tv.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        tv.setHeaderHidden(True)
        self.status_label = QLabel(self)
        layout = QVBoxLayout(self)
        layout.addWidget(self.status_label)
        layout.addWidget(self.tree_view)
        self.setLayout(layout)
        self.tree_view.setModel(available_fonts.text_fonts().misc_model())

        app.translateUI(self)

    def translateUI(self):
        self.status_label.setText(_("Fontconfig data:"))


class FontFilterProxyModel(QSortFilterProxyModel):
    """
    Custom proxy model.
    - Child elements are never filtered.
    - Font names that are also in the list of installed
      notation fonts are always filtered.
    """

    def filterAcceptsRow(self, row, parent):
        if parent.isValid():
            return True
        else:
            index = self.sourceModel().index(row, 0, parent)
            if index.data().lower() in _installed_notation_fonts:
                return False
            else:
                return super().filterAcceptsRow(row, parent)


class FontTreeModel(QStandardItemModel):
    """Custom Item Model holding information about available
    fonts. Builds the tree upon first use and caches the results
    until invalidated.
    Uses a custom filter mechanism to never filter child elements."""
    def __init__(self, parent=None):
        super().__init__(parent)
        self._proxy = FontFilterProxyModel(self)
        self._proxy.setSourceModel(self)

    def populate(self, families):
        """Populate the data model to be displayed in the results"""

        def sample(sub_family, style):
            """Produce a styled font sample for a given weight/style.
            This is not completely reliable since the style aliases
            returned by LilyPond are somewhat inconsistent and don't
            always match a font style that PyQt can get."""
            #L10N Please don't translate this 'sentence' literally but use an appropriate pangram in your language - or leave it untranslated
            item = QStandardItem(
                _('The quick brown fox jumps over the lazy dog')
            )
            font = QFont(sub_family)
            font.setStyleName(style)
            item.setFont(font)
            return item

        self.reset()
        family_names = sorted(families.keys(), key=lambda s: s.lower())
        root = self.invisibleRootItem()
        for name in family_names:
            family = families[name]
            sub_families = []
            for sub_family_name in sorted(family.keys()):
                sub_family = family[sub_family_name]
                if len(sub_family) == 1:
                    # Subfamily has only one entry, create single line
                    style = sub_family[0]
                    sub_families.append(
                        [QStandardItem('{} ({})'.format(
                            sub_family_name, style)),
                        sample(sub_family_name, style)])
                else:
                    # Subfamily has multiple entries, create
                    # container plus styled line for each style
                    sub_family_item = QStandardItem(sub_family_name)
                    sub_families.append(sub_family_item)
                    for style in sorted(sub_family):
                        sub_family_item.appendRow(
                            [QStandardItem(style),
                            sample(sub_family_name, style)])

            # Pull up subfamily as top-level entry if
            # - there is only one subfamily and
            # - it is a subfamily item with children
            if (len(sub_families) == 1
                and isinstance(sub_families[0], QStandardItem)
            ):
                root.appendRow(sub_families[0])
            else:
                family_item = QStandardItem(name)
                root.appendRow(family_item)
                for f in sub_families:
                    family_item.appendRow(f)

    def proxy(self):
        return self._proxy

    def reset(self):
        self.clear()
        self.setColumnCount(2)
        self.setHeaderData(0, Qt.Orientation.Horizontal, _("Font"))
        self.setHeaderData(1, Qt.Orientation.Horizontal, _("Sample"))


class MiscTreeModel(QStandardItemModel):
    """Custom tree model for miscellaneous fontconfig information.
    This model stores three top-level elements as individual fields
    and only then appends them to the root. So they are optionally
    available independently."""
    def __init__(self, parent=None):
        super().__init__(parent)
        self.reset()

    def populate(self, config_files, config_dirs, font_dirs):
        """Sort entries and construct the overall model."""
        self.reset()
        for file in sorted(config_files, key=lambda s: s.lower()):
            self.config_files.appendRow(QStandardItem(file))
        for config_dir in sorted(config_dirs, key=lambda s: s.lower()):
            self.config_dirs.appendRow(QStandardItem(config_dir))
        for font_dir in sorted(font_dirs, key=lambda s: s.lower()):
            self.font_dirs.appendRow(QStandardItem(font_dir))

        misc_root = self.invisibleRootItem()
        misc_root.appendRow(self.config_files)
        misc_root.appendRow(self.config_dirs)
        misc_root.appendRow(self.font_dirs)

    def reset(self):
        self.clear()
        self.config_files = QStandardItem(_("Configuration Files"))
        self.config_dirs = QStandardItem(_("Configuration Directories"))
        self.font_dirs = QStandardItem(_("Searched Font Directories"))


class TextFonts(QObject):
    """Provide information about available text fonts. These are exactly the
    fonts that can be seen by LilyPond.
    This is only produced upon request but then stored permanently during the
    program's runtime.
    load_fonts() will run LilyPond to determine the list of fonts, optionally
    reporting to a log.Log widget if given.
    Since this is an asynchronous process GUI elements that want to use the
    results have to connect to the 'loaded' signal which is emitted after
    LilyPond has completed and the results been parsed.

    A Fonts() object is immediately available as fonts.available_fonts, and
    its is_loaded member can be requested to test if fonts have already been
    loaded.
    """

    loaded = signals.Signal()

    def __init__(self, lilypond_info):
        super().__init__()
        self.lilypond_info = lilypond_info
        self._tree_model = FontTreeModel(self)
        self._misc_model = MiscTreeModel(self)
        self.job = None
        self.load_fonts()

    def reset(self, log_widget=None):
        self._log = []
        self._tree_model.reset()
        self._misc_model.reset()
        # all the functions we need are static members in Qt6
        self.font_db = QFontDatabase
        ## needs to be reset for the LilyPond-dependent fonts
        #self.font_db = QFontDatabase()

        self._is_loaded = False

    def log(self):
        return self._log

    def acknowledge_lily_fonts(self):
        """Add the OpenType fonts in LilyPond's font directory
        to Qt's font database. This should be relevant (untested)
        when the fonts are not additionally installed as system fonts."""
        # TODO: Move the the filtering here.
        # It's not correct to first add the notation fonts to the font debug
        # only to filter them again later. Besides, there might be other valid
        # fonts caught by the filter.
        font_dir = os.path.join(self.lilypond_info.datadir(), 'fonts', 'otf')
        for lily_font in os.listdir(font_dir):
            self.font_db.addApplicationFont(
                os.path.join(font_dir, lily_font)
            )

    def add_style_to_family(self, families, family_name, input):
        """Parse a font face definition provided by LilyPond.
        There is some guesswork involved since there may be
        discrepancies between the fonts/styles reported by
        LilyPond and those available in QFontDatabase.
        To discuss this the function is heavily commented.
        See also
        http://lists.gnu.org/archive/html/lilypond-user/2018-07/msg00338.html
        """

        def un_camel(style):
            """
            Try to 'fix' a class of errors when LilyPond reports
            e.g. 'BoldItalic' instead of 'Bold Italic'.
            It is unclear if this actually fixes the source of the
            issue or just one arbitrary example.
            """
            # The following regular expression would be a 'proper'
            # un-camel-ing, but this seems not to be relevant.
            #un_cameled = re.sub('([^ ])([A-Z][a-z])', r'\1 \2', style)
            #return re.sub('([A-Za-z])([0-9])', r'\1 \2', un_cameled)
            return (
                "Bold Italic" if style == "BoldItalic"
                else style
            )

        if family_name not in families.keys():
            families[family_name] = {}
        family = families[family_name]
        input = input.strip().split(':')
        # This is a safeguard against improper entries
        if len(input) == 2:
            # The first segment has always one or two entries:
            # - The font family name
            # - The font subfamily name if it differs from the base name.
            # Therefore the series is always the *last* entry in the list.
            # We "unescape" hyphens because this escape is not necessary
            # for our purposes.
            sub_family = input[0].split(',')[-1].replace('\\-', '-')
            if sub_family not in family.keys():
                family[sub_family] = []
            qt_styles = self.font_db.styles(sub_family)
            lily_styles = input[1][6:].split(',')
            match = ''
            if not qt_styles:
                # In some cases Qt does *not* report available styles.
                # In these cases it seems correct to use the style reported
                # by LilyPond. In very rare cases it seems possible that
                # LilyPond reports multiple styles for such fonts, and for now
                # we have to simply ignore these cases so we take the first
                # or single style.
                match = un_camel(lily_styles[0])
            else:
                # Match LilyPond's reported styles with those reported by Qt.
                # We need to un-camel the LilyPond-reported style name, but
                # this may not be a final fix (see comment to un_camel()).
                # If no match is found we simply hope that the style
                # reported by LilyPond will do.
                for style in lily_styles:
                    style = un_camel(style)
                    if style in qt_styles:
                        match = style
                        break
                if not match:
                    match = un_camel(lily_styles[0])
            if match not in family[sub_family]:
                family[sub_family].append(match)
        else:
            pass
            # TODO: issue a warning?
            # In my examples *some* fonts were parsed improperly
            # and therefore skipped.
            # I *think* this happens at the stage of splitting the
            # LilyPond log into individual lines.
            #print("Error when parsing font entry:")
            #print(name)
            #print(input)

    def flatten_log(self):
        """Flatten job history into flat string list."""
        for line in self.job.history():
            # lines in Job.history() are tuples of text and type,
            # we're only interested in the text.
            lines = line[0].split('\n')
            for l in lines:
                self._log.append(l)

    def is_loaded(self):
        return self._is_loaded

    def load_fonts(self, log_widget=None):
        """Run LilyPond to retrieve a list of available fonts.
        Afterwards process_results() will parse the output and build
        info structures to be used later.
        If a log.Log widget is passed as second argument this will
        be connected to the Job to provide realtime feedback on the process.
        Any caller should connect to the 'loaded' signal because this
        is an asynchronous task that takes long to complete."""
        self.reset()
        self.acknowledge_lily_fonts()
        self.run_lilypond(log_widget)

    def misc_model(self):
        return self._misc_model

    def model(self):
        return self._tree_model

    def parse_entries(self):
        """Parse the LilyPond log and push entries to the various
        lists and dictionaries. Parsing the actual font style
        definition is deferred to add_style_to_family()."""
        regexp = re.compile(r'(.*)\\-\d+')
        families = {}
        config_files = []
        config_dirs = []
        font_dirs = []
        last_family = None
        for e in self._log:
            if e.startswith('family'):
                # NOTE: output of this process is always English,
                # so we can hardcode the splice indices
                original_family = e[7:]
                # filter size-indexed font families
                basename = regexp.match(original_family)
                last_family = basename.groups()[0] if basename else original_family
            elif last_family:
                # We're in the second line of a style definition
                if not last_family.endswith('-brace'):
                    self.add_style_to_family(families, last_family, e)
                last_family = None
            elif e.startswith('Config files:'):
                config_files.append(e[14:])
            elif e.startswith('Font dir:'):
                font_dirs.append(e[10:])
            elif e.startswith('Config dir:'):
                config_dirs.append(e[12:])

        return families, config_files, config_dirs, font_dirs

    def process_results(self):
        """Parse the job history list to dictionaries."""

        self.flatten_log()
        families, config_files, config_dirs, font_dirs = self.parse_entries()

        self._tree_model.populate(families)
        self._misc_model.populate(config_files, config_dirs, font_dirs)

        self._is_loaded = True
        self.job = None
        self.loaded.emit()

    def run_lilypond(self, log_widget=None):
        """Run lilypond from info with the args list, and a job title."""
        # TODO: Use the global JobQueue
        info = self.lilypond_info
        j = self.job = job.Job(
            [info.abscommand() or info.command] + ['-dshow-available-fonts'])
        j.set_title(_("Available Fonts"))
        j.done.connect(self.process_results)
        if log_widget:
            log_widget.connectJob(j)
        j.start()



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2020 - 2020 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
This module is imported from the startup script to perform some checks whether
Frescobaldi can run properly.

The user-visible strings need not to be marked for translation, as the
translation infrastructure is not yet set up.

"""

import os
import importlib.util
import sys

import appinfo


def error(title, message, can_continue=False):
    """Display an error, and exit if can_continue is False.

    Displays a message box if PyQt6 is available, otherwise write the error to
    stderr.

    """
    try:
        from PyQt6.QtWidgets import QApplication, QMessageBox
    except ImportError:
        write = sys.stderr.write
        write(title + '\n')
        write(message + '\n')
        if can_continue:
            write("Continuing...\n")
        else:
            write("Exiting.\n")
            sys.exit(1)
    else:
        app = QApplication([])
        box = QMessageBox()
        box.setIcon(QMessageBox.Icon.Critical)
        box.setWindowTitle("Frescobaldi")
        box.setText(title)
        box.setInformativeText(message)
        if can_continue:
            box.setStandardButtons(QMessageBox.StandardButton.Abort|QMessageBox.StandardButton.Ignore)
            box.button(QMessageBox.StandardButton.Ignore).setText("Continue")
        else:
            box.setStandardButtons(QMessageBox.StandardButton.Abort)
        box.button(QMessageBox.StandardButton.Abort).setText("Quit")
        result = box.exec()
        app.quit()
        if not can_continue or result == QMessageBox.StandardButton.Abort:
            sys.exit(1)


# Check Python version
v = sys.version_info
r = appinfo.required_python_version
if v < r:
    error("Python version too old",
        f"Frescobaldi is started with Python {v[0]}.{v[1]} \
but requires at least version {r[0]}.{r[1]}.")

# Check Qt version
from PyQt6.QtCore import QLibraryInfo, QVersionNumber
qt_version = QLibraryInfo.version()
## It's better to test for specific features than the overall Qt version
#qt_required = QVersionNumber(*appinfo.required_qt_version)
#if qt_version < qt_required:
#    error("Qt version is too old.",
#        f"Frescobaldi is started with Qt {qt_version.toString()} \
#but requires at least version {qt_required.toString()}.")

# Check QtSvg availability
if importlib.util.find_spec('PyQt6.QtSvg') is None:
    error("Missing Qt module",
        "Frescobaldi cannot start because the "
        "PyQt6.QtSvg module is not installed.")

# Check QtPdf availability
if importlib.util.find_spec('PyQt6.QtPdf') is None:
    error("Missing Qt module",
        "Frescobaldi cannot start because the "
        "PyQt6.QtPdf module is not installed.")

# Check qpageview availability
if importlib.util.find_spec('qpageview') is None:
    error("The 'qpageview' module can't be found.",
        "Frescobaldi can't find the 'qpageview' module. This module is "
        "required for the Music View and other viewers inside Frescobaldi, "
        "and can be downloaded from http://github.com/frescobaldi/qpageview."
        "\n\n"
        "Unfortunately, Frescobaldi cannot run without it.")

# Check for PDF link support (added in Qt 6.6)
# As of 2026, some Linux distros still ship older Qt versions without it.
try:
    from PyQt6.QtPdf import QPdfLinkModel
except ImportError:
    error("Point-and-click unavailable",
        f"You are using an older Qt version ({qt_version.toString()}) "
        "that lacks support for links in PDF documents. "
        "Frescobaldi will run, but point-and-click will be unavailable."
        "\n\n"
        "To fix this, install Qt 6.6 or newer.",
        can_continue=True)


# The python-ly check could also move where, but we don't do it now, because
# the --python-ly option could point to another path, and the startup arguments
# are parsed later.



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.


"""
Manage lists of music/notation fonts, providing tools to install or
uninstall to/from a LilyPond installation."
"""


import os
import re
from enum import Enum
from shutil import copyfile
from pathlib import Path

from PyQt6.QtCore import (
    QObject,
    QSettings,
    Qt
)
from PyQt6.QtGui import (
    QStandardItem,
    QStandardItemModel,
)
from PyQt6.QtWidgets import (
    QAbstractItemView,
    QFileDialog,
    QHBoxLayout,
    QMessageBox,
    QPushButton,
    QTreeView,
    QVBoxLayout,
    QWidget,
)

import app
import fonts
from . import textfonts


class MusicFontsWidget(QWidget):
    """Display list of installed music fonts,
    show font preview score, install/remove fonts."""

    def __init__(self, parent):
        super().__init__(parent)
        s = QSettings()
        s.beginGroup('music-fonts')
        self._auto_install = s.value('auto-install', True, bool)
        self._font_repo = fonts.music_fonts_repo()

        self.music_fonts = parent.available_fonts.music_fonts()

        layout = QVBoxLayout()
        self.setLayout(layout)

        self.button_download = bd = QPushButton(self)
        # TODO: This hasn't been implemented yet
        bd.setEnabled(False)
        self.button_auto_install = bai = QPushButton(self)
        self.button_install = bi = QPushButton(self)
        self.button_remove = br = QPushButton(self)
        bai.setEnabled(bool(self._font_repo and not self._auto_install))
        br.setEnabled(False)
        self.button_download.clicked.connect(self.download_music_fonts)
        self.button_auto_install.clicked.connect(self.install_button_clicked)
        self.button_install.clicked.connect(self.install_button_clicked)
        self.button_remove.clicked.connect(self.remove_music_font)
        bl = QHBoxLayout()
        bl.addStretch()
        bl.addWidget(bai)
        bl.addWidget(bi)
        bl.addWidget(bd)
        bl.addWidget(br)
        layout.addLayout(bl)

        # If conditions for automatic install are met do so
        if self._font_repo and self._auto_install:
            self.install_music_fonts('auto')

        self.tree_view = tv = QTreeView(self)
        tv.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        tv.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        tv.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        tv.setModel(self.music_fonts.item_model())
        tv.selectionModel().selectionChanged.connect(
            self.music_fonts_selection_changed
        )
        layout.addWidget(tv)

        app.translateUI(self)

    def translateUI(self):
        self.button_remove.setText(_("Remove..."))
        self.button_remove.setToolTip(_("Remove selected music font"))
        self.button_download.setText(_("Download..."))
        self.button_download.setToolTip(_(
            "Download music fonts from a repository on Github.\n"
            "NOTE: Not implemented yet."
        ))
        self.button_install.setText(_("Install..."))
        self.button_install.setToolTip(_(
            "Link fonts from a directory to the current LilyPond installation"
        ))
        self.button_auto_install.setText(_("Install (repo)"))
        self.button_auto_install.setToolTip(_(
            "Link fonts from the global music font repository\n"
            "to the current LilyPond installation."
        ))

    def download_music_fonts(self):
        """Opens a dialog to download and install music fonts from Github."""
        # TODO: implement
        # from . import download
        # ...
        pass

    def install_button_clicked(self):
        mode = 'auto' if self.sender() == self.button_auto_install else 'dlg'
        self.install_music_fonts(mode)

    def install_music_fonts(self, mode):
        """'Install' music fonts from a directory (structure) by
        linking fonts into the LilyPond installation's font
        directories (otf and svg)."""

        if mode == 'auto':
            # Use the global fonts repository
            repo = fonts.music_fonts_repo()
        else:
            dlg = QFileDialog(self)
            dlg.setFileMode(QFileDialog.FileMode.Directory)
            if not dlg.exec():
                return
            repo = MusicFontRepo(dlg.selectedFiles()[0])

        installed = self.music_fonts
        repo.flag_for_install(installed)

        try:
            repo.install_flagged(installed)
            if mode == 'auto':
                self.button_auto_install.setEnabled(False)
        except MusicFontPermissionException as e:
            msg_box = QMessageBox()
            msg_box.setText(_("Fonts could not be installed!"))
            msg_box.setInformativeText(_(
                "Installing fonts in the LilyPond installation "
                "appears to require administrator privileges on "
                "your system and can unfortunately not be handled "
                "by Frescobaldi."
            ))
            msg_box.setDetailedText(f"{e}")
            msg_box.exec()

    def music_font_family(self):
        try:
            family_name = (
                self.tree_view.selectionModel().selectedIndexes()[0].data())
        except IndexError:
            family_name = 'emmentaler'
        return family_name

    def music_fonts_selection_changed(self, new, old):
        """Show a new score example with the selected music font"""
        has_font = self.tree_view.selectionModel().hasSelection()
        if has_font:
            music_name = self.music_font_family()
            brace_name = (
                music_name
                if self.music_fonts.family(music_name).has_brace('otf')
                else 'emmentaler'
            )
            self.window().select_font('music', music_name)
            self.window().select_font('brace', brace_name)
        else:
            self.window().select_font('emmentaler')
            self.window().select_font('emmentaler')
        self.window().invalidate_command()
        self.button_remove.setEnabled(has_font)

    def remove_music_font(self):
        """Remove one or more font family/ies from the LilyPond installation.
        Works only for *links*, not for *files*."""
        text = ''
        informative_text = ''
        detailed_text = ''
        try:
            indexes = self.tree_view.selectionModel().selectedRows()
            self.music_fonts.remove(indexes)
            self.button_auto_install.setEnabled(bool(self._font_repo))
        except MusicFontFileRemoveException as e:
            text = _("Font family could not be removed!")
            informative_text = _(
                "To avoid persistent damage Frescobaldi only supports "
                + "removing music fonts that are linked into a LilyPond "
                + "installation. The font being removed includes real "
                + "files and can therefore not be removed directly.")
            detailed_text = f"{e}"
        if text:
            from PyQt6.QtWidgets import QMessageBox
            msg_box = QMessageBox()
            msg_box.setText(text)
            msg_box.setInformativeText(informative_text)
            msg_box.setDetailedText(detailed_text)
            msg_box.exec()


class MusicFontException(Exception):
    pass


class MusicFontPermissionException(MusicFontException):
    pass


class MusicFontFileRemoveException(MusicFontException):
    pass


class MusicFontStatus(Enum):
    """Status value enumeration for registered font files.
    Actually the 'MISSING' and 'BROKEN' values should never
    occur since this is already checked upon the creation
    of any font object."""
    FILE = 0
    MISSING_FILE = 1
    LINK = 2
    BROKEN_LINK = 3
    MISSING = 4


class MusicFontFile(QObject):
    """Represents a font file within a font family."""

    def __init__(self, file):
        self.file = file
        self.status = None
        self.install = False


class MusicFontFamily(QObject):
    """Represents a single music font family.
    Keep track of completeness status and add files to the family.
    Other classes make use of some class functions and variables."""

    # List of sizes expected for a complete music font
    sizes_list = ['11', '13', '14', '16', '18', '20', '23', '26']
    # Regular expression to determine a LilyPond music font
    font_re = re.compile(
        r'(?P<family>.*)-(?P<size>brace|\d\d)\.(?P<type>otf|svg|woff)$')

    @classmethod
    def parse_filename(cls, file):
        """Check if a given filename represents a LilyPond music font.
        Returns a match object with three named groups ('family', 'size',
        'type') or None if no match is found."""
        return MusicFontFamily.font_re.match(os.path.basename(file))

    @classmethod
    def check_file(cls, file):
        """Test if a file is a LilyPond music font.
        Returns three family, type, size if successful or raises a
        MusicFontException otherwise."""
        if not os.path.exists(file):
            raise MusicFontException(
                f'Not an existing file or link: {file}')
        font = cls.parse_filename(file)
        if not font:
            raise MusicFontException(
                f'File {file} does not appear to be a valid font file')
        return font['family'], font['type'], font['size']

    def __init__(self, file=None):
        self.family = None
        self._files = {
            'otf': {},
            'svg': {},
            'woff': {}
        }
        if file:
            self.add_file(file)

    def __getitem__(self, key):
        return self._files[key]

    def add(self, type, size, file):
        """Add a file if it has already been parsed to type/size.
        Existing entries are silently overwritten."""
        self._files[type][size] = MusicFontFile(file)

    def add_file(self, file):
        """Add a given file to the font family if it is a valid
        music font and if it does not belong to a different font
        family."""
        family, type, size = MusicFontFamily.check_file(file)
        if self.family and self.family != family:
            raise MusicFontException(
                'File {name} does not belong to font family {family}'.format(
                    name=file, family=self.family))
        if not self.family:
            self.family = family
        self.add(type, size, file)

    def flag_all_for_install(self):
        for type in self._files:
            for size in self._files[type]:
                self._files[type][size].install = True

    def flag_for_install(self, target_family):
        for type in self._files:
            for size in self._files[type]:
                file = self._files[type][size]
                if (
                    self.status(type, size)
                    in [MusicFontStatus.FILE, MusicFontStatus.LINK]
                    and target_family.status(type, size)
                    not in [MusicFontStatus.FILE, MusicFontStatus.LINK]
                ):
                    file.install = True

    def has_brace(self, type):
        return 'brace' in self._files[type].keys()

    def is_complete(self, type=None):
        """Determines if the given type is complete with all sizes
        and a brace font. If no type is given *all* types are checked."""
        if type:
            return self.has_brace(type) and not self.missing_sizes(type)
        else:
            return (
                self.is_complete('otf')
                and self.is_complete('svg')
                and self.is_complete('woff')
            )

    def missing_sizes(self, type):
        """Returns a list of size strings representing missing
        font sizes for the given type. For a complete font this
        will be an empty list, which is checked in is_complete()
        for example."""
        return (
            [size for size in MusicFontFamily.sizes_list
                if size not in self.sizes(type)]
        )

    def remove(self, type, size):
        """Remove a given type/size combination."""
        self._files[type].pop(size, None)

    def sizes(self, type):
        """Returns a string list with the installed sizes for a given type."""
        return sorted(self._files[type].keys())

    def status(self, type, size):
        """Returns the status for a given type/size combination."""
        if size not in self._files[type].keys():
            return MusicFontStatus.MISSING

        font = self._files[type][size]
        if font.status is None:
            from pathlib import Path
            file = Path(font.file)
            if file.is_symlink():
                if file.exists():
                    font.status = MusicFontStatus.LINK
                else:
                    font.status = MusicFontStatus.BROKEN_LINK
            elif file.is_file():
                font.status = MusicFontStatus.FILE
            else:
                font.status = MusicFontStatus.MISSING_FILE
        return font.status

    def walk(self):
        """Return entries for all registered fonts."""
        for type in self._files:
            for size in self._files[type]:
                yield type, size, self._files[type][size]


class AbstractMusicFontList(QObject):
    """Abstract class managing a list of music fonts."""

    def __init__(self):
        self._item_model = None
        self.clear()

    def add_file(self, file):
        """Add the given file to a MusicFontFamily.
        If the file doesn't point to a music font a MusicFontException
        will simply be forwarded, so a caller must handle that.
        If the font family is already present the file is added to it
        (silently overwriting an existing type/size combination),
        otherwise the MusicFont object is created."""
        family, type, size = MusicFontFamily.check_file(file)
        if family not in self._families.keys():
            self._families[family] = MusicFontFamily(file)
        else:
            self._families[family].add(type, size, file)

    def add_family(self, family):
        """Add a MusicFontFamily object that is already composed.
        An existing family with the same name will get overwritten."""
        family_name = family.family
        self._families[family_name] = family

    def clear(self):
        """Reset the list of fonts/files."""
        self._families = {}
        if self._item_model:
            self._item_model.reset()

    def families(self):
        """Return a sorted list with all family names."""
        return sorted(self._families.keys())

    def family(self, name):
        """Returns a MusicFont for the given family name,
        or None if it isn't present yet."""
        return self._families.get(name, None)

    def add_tree(self, root):
        """Walk through a given directory tree and add all found
        music fonts to the list."""
        for dir, dirs, files in os.walk(root):
            for file in files:
                try:
                    self.add_file(os.path.join(dir, file))
                except Exception:
                    # file is not a music font, ignore
                    pass

    def item_model(self):
        """Return the (cached) item model."""
        if not self._item_model:
            self._item_model = MusicFontsModel()
            self._item_model.populate(self)
        return self._item_model

    def walk(self):
        """Generator to produce *all* font files for the repository."""
        for family_name in self._families:
            family = self.family(family_name)
            for type in family._files:
                for size in family._files[type]:
                    yield (
                        family, family_name, type, size,
                        family._files[type][size]
                    )


class MusicFontRepo(AbstractMusicFontList):
    """Represents a repository of music fonts, typically within a
    single directory tree."""

    def __init__(self, root):
        super().__init__()
        self._root = root
        self.installable_fonts = AbstractMusicFontList()
        self.add_tree(root)

    def flag_for_install(self, installed):
        """Determine which fonts can be installed."""
        self.installable_fonts.clear()
        for family_name in self._families:
            repo_family = self.family(family_name)
            target_family = installed.family(family_name)
            if not target_family:
                repo_family.flag_all_for_install()
            else:
                repo_family.flag_for_install(target_family)
        for _, _, _, _, font in self.walk():
            if font.install:
                self.installable_fonts.add_file(font.file)

    def install_flagged(self, target):
        """Install all font files in the target repo that have
        been flagged for installation."""
        for _, _, type, _, font in self.installable_fonts.walk():
            target.install(type, font.file)

        target.item_model().populate(target)

    def root(self):
        return self._root


class InstalledMusicFonts(AbstractMusicFontList):
    """Manages the music fonts installed in a given LilyPond
    installation. Provides means to add/remove fonts if the
    file system permissions allow to do so."""

    def __init__(self, lilypond_info):
        super().__init__()
        self.lilypond_info = lilypond_info
        self.font_root = os.path.join(lilypond_info.datadir(), 'fonts')
        self.add_tree(self.font_root)

    def font_dir(self, type):
        """Return the font directory for the given type.
        SVG and WOFF share a directory."""
        last_segment = 'otf' if type == 'otf' else 'svg'
        return os.path.join(self.font_root, last_segment)

    def install(self, type, font_file, copy=False):
        """Install a font file in the type's directory.
        Raise an exception if this fails (typically
        lacking permissions)."""
        target = os.path.join(
            self.font_root,
            self.font_dir(type),
            os.path.basename(font_file)
        )
        if copy:
            try:
                copyfile(font_file, target)
            except OSError as e:
                raise MusicFontPermissionException(
                    _("Font installation failed:") + f"\n{e}")
        else:
            try:
                os.symlink(font_file, target)
            except NotImplementedError:
                # On Windows prior to Vista symlinks are not supported
                self.install(type, font_file, copy=True)
            except OSError as e:
                raise MusicFontPermissionException(
                    _("Font installation failed:") + f"\n{e}")
        self.add_file(target)

    def remove(self, indexes):
        """Remove one or more font family/ies from the LilyPond installation.
        If any of the fonts includes any real files (as opposed to links)
        abort without removing *anything* and raise an exception."""
        for index in indexes:
            family_name = index.data()
            family = self.family(family_name)
            links = []
            files = []
            for type, size, font in family.walk():
                file = Path(font.file)
                if file.is_symlink():
                    links.append(file)
                else:
                    files.append(str(file))
            if files:
                raise MusicFontFileRemoveException(_("\n".join(files)))
            try:
                for link in links:
                    link.unlink()
            except OSError as e:
                raise MusicFontPermissionException(
                    _("Font removal failed:") + f"\n{e}")
            self._families.pop(family_name, None)
        self.item_model().populate(self)


class MusicFontsModel(QStandardItemModel):
    """Data model to maintain the list of available music fonts."""

    def populate(self, fonts):
        """Populate the data model from the fonts dictionary."""

        def check_type(font, type):
            size_result = QStandardItem()
            sizes = font.missing_sizes(type)
            if sizes:
                if len(sizes) == 8:
                    size_result.setCheckState(Qt.CheckState.Unchecked)
                else:
                    size_result.setCheckState(Qt.CheckState.PartiallyChecked)
                    size_result.setText(_("Missing:") + " {}".format(
                        ", ".join(sizes)))
            else:
                size_result.setCheckState(Qt.CheckState.Checked)

            has_brace = Qt.CheckState.Checked if font.has_brace(type) else Qt.CheckState.Unchecked
            brace_result = QStandardItem()
            brace_result.setCheckState(has_brace)
            return [size_result, brace_result]

        self.reset()
        textfonts._installed_notation_fonts = installed = []
        for family_name in fonts.families():
            installed.append(family_name.lower())
            font = fonts.family(family_name)
            result = [QStandardItem(family_name)]
            result.extend(check_type(font, 'otf'))
            result.extend(check_type(font, 'svg'))
            result.extend(check_type(font, 'woff'))
            self.invisibleRootItem().appendRow(result)

    def reset(self):
        self.clear()
        self.setColumnCount(7)
        self.setHeaderData(0, Qt.Orientation.Horizontal, _("Font"))
        self.setHeaderData(1, Qt.Orientation.Horizontal, _("OpenType"))
        self.setHeaderData(2, Qt.Orientation.Horizontal, _("(Brace)"))
        self.setHeaderData(3, Qt.Orientation.Horizontal, _("SVG"))
        self.setHeaderData(4, Qt.Orientation.Horizontal, _("(Brace)"))
        self.setHeaderData(5, Qt.Orientation.Horizontal, _("WOFF"))
        self.setHeaderData(6, Qt.Orientation.Horizontal, _("(Brace)"))
